/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>  	/* for read/write to files */
#include "m_imp.h"
#include "g_canvas.h"

t_class *array_class;

struct _array
{
    t_gobj x_gobj;
    t_glist *x_glist;
    int x_n;
    int x_elemsize;
    char *x_vec;
    t_symbol *x_name;
    t_symbol *x_template;
    t_float x_firstx;	    /* X value of first item */
    t_float x_xinc; 	    /* X increment */
};


void graph_array(t_glist *gl, t_symbol *s, t_intarg n, t_symbol *template)
{
    static int gcount = 0;
     int zz;
   t_array *x;
    t_canvas *c;
    char *str;
    if (s == &s_)
    {
    	char buf[40];
    	sprintf(buf, "array%d", ++gcount);
    	s = gensym(buf);    	
    	template = &s_float;
    	n = 100;
    }
    else if (!strncmp((str = s->s_name), "array", 5)
    	&& (zz = atoi(str + 5)) > gcount) gcount = zz;
    c = (t_canvas *)pd_findbyclass(template, canvas_class);
    if (!c && template != &s_float)
    {
    	error("array: couldn't find template %s", template->s_name);
    }
    x = (t_array *)pd_new(array_class);
    
    pd_bind(&x->x_gobj.g_pd, s);
    if (n <= 0) n = 100;  
    x->x_n = n;
    x->x_elemsize = canvas_template_size(c);
    x->x_vec = getbytes(x->x_n * x->x_elemsize);
    memset(x->x_vec, 0, x->x_n * x->x_elemsize);
    	/* LATER should check that malloc */
    x->x_name = s;
    x->x_template = template;
    x->x_firstx = 0;
    x->x_xinc = 1;	    	/* LATER make methods to set this... */
    glist_add(gl, &x->x_gobj);
    x->x_glist = gl;
}

void canvas_array(t_glist *canvas)
{
    t_glist *gl;
    gl = glist_findgraph(canvas, 0, 100, -1, 1);
    graph_array(gl, &s_, 0, 0);
}

static void array_free(t_array *x)
{
    pd_unbind(&x->x_gobj.g_pd, x->x_name);
    freebytes(x->x_vec, x->x_n * x->x_elemsize);
}

/* -------------------- widget behavior for array ------------ */

static void array_getrect(t_gobj *z, t_glist *owner,
    int *x1, int *y1, int *x2, int *y2)
{
    t_array *x = (t_array *)z;
    *x1 = 0;
    *y1 = 0;
    *x2 = 0;
    *y2 = 0; /* fill in later */ 
}

static void array_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    /* refuse */
}

static void array_select(t_gobj *z, t_glist *glist, int state)
{
    t_array *x = (t_array *)z;
    /* fill in later */
}

static void array_activate(t_gobj *z, t_glist *glist, int state)
{
    post("array_activate %d", state);
}

static void array_delete(t_gobj *z, t_glist *glist)
{
    /* nothing to do */
}

static void array_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_array *x = (t_array *)z;
    if (vis)
    {
    	int i, elemsize, xonset, yonset;
    	t_field *fx, *fy;
    	t_canvas *c = (t_canvas *)pd_findbyclass(x->x_template, canvas_class);
    	elemsize = canvas_template_size(c);
    	if (!(fy = canvas_find_field(c, gensym("y"), &yonset))
    	    || field_type(fy) != &s_float)
    	{
    	    error("%s: needs floating-point 'y' field", x->x_template->s_name);
    	    sys_vgui(".x%x.c create text 50 50 -text foo\
    	    	-tags .x%x.a%x\n",
    	    	glist_getcanvas(glist), glist_getcanvas(glist), x);
    	}
    	else if (!(fx = canvas_find_field(c, gensym("x"), &xonset))
    	    || field_type(fx) != &s_float)
    	{
    	    float firsty, xcum = x->x_firstx;
    	    int lastpixel = -1, ndrawn = 0;
    	    sys_vgui(".x%x.c create line \\\n", glist_getcanvas(glist));
    	    for (i = 0; i < x->x_n; i++)
    	    {
    	    	float yval = *(float *)((x->x_vec + elemsize * i) + yonset);
    	    	float xpix = glist_xtopixels(glist, xcum);
    	    	int ixpix = xpix + 0.5;
    	    	if (ixpix != lastpixel)
    	    	{
    	    	    sys_vgui("%d %f \\\n", ixpix,
    	    	    	glist_ytopixels(glist, yval));
    	    	    ndrawn++;
    	    	}
    	    	lastpixel = ixpix;
    	    	if (ndrawn >= 1000) break;
    	    	xcum += x->x_xinc;
    	    }
    	    sys_vgui("-tags .x%x.a%x\n", glist_getcanvas(glist), x);
    	    firsty = *(float *)(x->x_vec + yonset);
    	    sys_vgui(".x%x.c create text %f %f -text {%s} -anchor e\
    	    	 -font %s -tags .x%x.a%x\n",
    	    	glist_getcanvas(glist),
    	    	glist_xtopixels(glist, x->x_firstx) - 5.,
    	    	glist_ytopixels(glist, firsty),
    	    	x->x_name->s_name, glist_getfont(glist),
    	    	    glist_getcanvas(glist), x);
    	}
    	else
    	{
    	    post("x, y arrays not yet supported");
    	}
    }
    else
    {
    	sys_vgui(".x%x.c delete .x%x.a%x\n",
    	    glist_getcanvas(glist), glist_getcanvas(glist), x);
    }
}

static void array_save(t_gobj *z, t_binbuf *b)
{
    t_array *x = (t_array *)z;
    binbuf_addv(b, "sssis;", gensym("#X"), gensym("array"),
    	x->x_name, x->x_n, x->x_template);
}

t_widgetbehavior array_widgetbehavior =
{
    array_getrect,
    array_displace,
    array_select,
    array_activate,
    array_delete,
    array_vis,
    array_save
};

void array_redraw(t_array *x)
{
    if (glist_isvisible(x->x_glist))
    {
    	array_vis(&x->x_gobj, x->x_glist, 0); 
    	array_vis(&x->x_gobj, x->x_glist, 1);
    }
}

t_canvas *array_template(t_array *x)
{
    return ((t_canvas *)pd_findbyclass(x->x_template, canvas_class));
}

int array_npoints(t_array *x)
{
    return (x->x_n);
}

char *array_vec(t_array *x)
{
    return (x->x_vec);
}

/*------------------- Pd messages ------------------------ */
static void array_const(t_array *x, double g)
{
    int i, elemsize, yonset;
    t_field *fy;
    t_canvas *c = (t_canvas *)pd_findbyclass(x->x_template, canvas_class);
    elemsize = canvas_template_size(c);
    if (!(fy = canvas_find_field(c, gensym("y"), &yonset))
    	|| field_type(fy) != &s_float)
    	    error("%s: needs floating-point 'y' field", x->x_template->s_name);
    else for (i = 0; i < x->x_n; i++)
    	*(float *)((x->x_vec + elemsize * i) + yonset) = g;
    array_redraw(x);
}

static void array_read(t_array *x, t_symbol *filename)
{
    int i, elemsize, yonset;
    t_field *fy;
    FILE *fd;
    t_canvas *c = (t_canvas *)pd_findbyclass(x->x_template, canvas_class);
    elemsize = canvas_template_size(c);
    if (!(fy = canvas_find_field(c, gensym("y"), &yonset))
    	|| field_type(fy) != &s_float)
    {
    	error("%s: needs floating-point 'y' field", x->x_template->s_name);
    	return;
    }
    if (!(fd = fopen(filename->s_name, "r")))
    {
    	error("%s: can't open", filename->s_name);
    	return;
    }
    for (i = 0; i < x->x_n; i++)
    {
    	if (!fscanf(fd, "%f", (float *)((x->x_vec + elemsize * i) + yonset)))
    	{
    	    post("%s: only %d elements out of %d read",
    	    	filename->s_name, i, x->x_n);
    	    break;
    	}
    }
    fclose(fd);
    array_redraw(x);
}

static void array_write(t_array *x, t_symbol *filename)
{
    int i, elemsize, yonset;
    t_field *fy;
    FILE *fd;
    t_canvas *c = (t_canvas *)pd_findbyclass(x->x_template, canvas_class);
    elemsize = canvas_template_size(c);
    if (!(fy = canvas_find_field(c, gensym("y"), &yonset))
    	|| field_type(fy) != &s_float)
    {
    	error("%s: needs floating-point 'y' field", x->x_template->s_name);
    	return;
    }
    if (!(fd = fopen(filename->s_name, "w")))
    {
    	error("%s: can't create", filename->s_name);
    	return;
    }
    for (i = 0; i < x->x_n; i++)
    {
    	if (!fprintf(fd, "%g", *(float *)((x->x_vec + elemsize * i) + yonset)))
    	{
    	    post("%s: write error", filename->s_name);
    	    break;
    	}
    }
    fclose(fd);
}

static void array_resize(t_array *x, t_intarg n)
{
    int was = x->x_n, elemsize;
    t_glist *gl;
    t_canvas *c;

    if (n < 0) n = 0;
    c = (t_canvas *)pd_findbyclass(x->x_template, canvas_class);
    elemsize = canvas_template_size(c);
    x->x_vec = t_resizebytes(x->x_vec, was * elemsize, n * elemsize);
    	/* LATER should check t_resizebytes result */
    if (n > was)
    	memset(x->x_vec + was*elemsize,
    	    0, (n - was) * elemsize);
    x->x_n = n;
    
    	/* if this is the only array in the graph,
    	    reset the graph's coordinates */
    gl = x->x_glist;
    if (gl->gl_list == &x->x_gobj && !x->x_gobj.g_next)
    {
    	vmess(&gl->gl_pd, gensym("bounds"), "ffff",
    	    0., gl->gl_y1, (double)n, gl->gl_y2);
    }
    else array_redraw(x);
    
}

static void array_print(t_array *x)
{  
    post("array %s: template %s, length %d",
    	x->x_name->s_name, x->x_template->s_name, x->x_n);
}

/*------------------- access functions at C level ---------------------*/
float array_get(t_array *x, t_symbol *s, t_int index)
{
    int elemsize, yonset;
    t_field *fy;
    t_canvas *c = (t_canvas *)pd_findbyclass(x->x_template, canvas_class);
    elemsize = canvas_template_size(c);
    if (!(fy = canvas_find_field(c, s, &yonset))
    	|| field_type(fy) != &s_float)
    {
    	error("%s: needs floating-point '%s' field", x->x_template->s_name,
    	    s->s_name);
    	return (0);
    }
    if (index < 0) index = 0;
    else if (index >= x->x_n) index = x->x_n - 1;
    return (*(float *)((x->x_vec + elemsize * index) + yonset));
}

void g_array_setup()
{
    array_class = class_new(gensym("array"), 0, array_free, sizeof(t_array),
    	CLASS_GOBJ, A_DEFSYM, A_DEFINT, A_DEFSYM, 0);
    class_setwidget(array_class, &array_widgetbehavior);
    class_addmethod(array_class, array_const, gensym("const"),
    	A_DEFFLOAT, A_NULL);
    class_addmethod(array_class, array_read, gensym("read"),
    	A_SYMBOL, A_NULL);
    class_addmethod(array_class, array_write, gensym("write"),
    	A_SYMBOL, A_NULL);
    class_addmethod(array_class, array_resize, gensym("resize"),
    	A_INT, A_NULL);
    class_addmethod(array_class, array_print, gensym("print"), A_NULL);
}

