/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

#include <stdlib.h>
#include <stdio.h>
#include "m_imp.h"
#include "g_canvas.h"
#include <string.h>

/* LATER:
  how can we check that the font exists?
*/

typedef struct _canvasenvironment
{
    t_symbol *ce_dir;
    int ce_argc;
    t_atom *ce_argv;
} t_canvasenvironment;

struct _canvas
{
    t_glist x_glist;
    t_symbol *x_font;
    struct _subcanvas *x_parent;
    struct _canvas *x_next;
    t_canvasenvironment *x_env;
    unsigned int x_dirty:1;
    unsigned int x_lastmoved:1;
    unsigned int x_vis:1;
    unsigned int x_willvis:1;
    unsigned int x_edit:1;
    unsigned int x_loading:1;
};

#define x_pd x_glist.gl_gobj.g_pd
#define x_gobj x_glist.gl_gobj
#define x_list x_glist.gl_list

typedef struct _subcanvas
{
    t_object x_ob;
    t_canvas *x_subcanvas;
    t_canvas *x_supercanvas;
} t_subcanvas;

/* ---------------------- variables --------------------------- */

t_class *canvas_class;
static t_class *subcanvas_class;

static t_canvas *canvas_list;	/* list of all visible canvases */
static int canvas_dspstate;

static t_symbol *canvas_newfilename = &s_;
static t_symbol *canvas_newdirectory = &s_;
static int canvas_newargc;
static t_atom *canvas_newargv;

t_canvas *canvas_editing;    /* last canvas to start text edting */ 

/* ------------------ forward function declarations --------------- */
void canvas_start_dsp(void);
void canvas_stop_dsp(void);
int canvas_getdspstate(void);

static void canvas_drawlines(t_canvas *x);

/* --------- functions to handle the canvas environment ----------- */

void canvas_setargs(int argc, t_atom *argv)
{
    if (canvas_newargv) bug("canvas_setargs");
    canvas_newargc = argc;
    canvas_newargv = copybytes(argv, argc * sizeof(t_atom));
}

void glob_setfilename(void *dummy, t_symbol *filename, t_symbol *dirname)
{
    canvas_newfilename = filename;
    canvas_newdirectory = dirname;
}

t_canvas *canvas_getcurrent(void)
{
    return ((t_canvas *)pd_findbyclass(&s__X, canvas_class));
}

t_canvas *canvas_setcurrent(t_canvas *x)
{
    pd_pushsym(&x->x_glist.gl_pd);
}

t_canvas *canvas_unsetcurrent(t_canvas *x)
{
    pd_popsym(&x->x_glist.gl_pd);
}

static t_canvasenvironment *canvas_getenv(t_canvas *x)
{
    if (!x) bug("canvas_getenv");
    while (!x->x_env)
    	if (!(x = x->x_parent->x_supercanvas))
    	    bug("t_canvasenvironment");
    return (x->x_env);
}

void canvas_getargs(int *argcp, t_atom **argvp)
{
    t_canvasenvironment *e = canvas_getenv(canvas_getcurrent());
    *argcp = e->ce_argc;
    *argvp = e->ce_argv;
}

t_symbol *canvas_getcurrentdir(void)
{
    t_canvasenvironment *e = canvas_getenv(canvas_getcurrent());
    return (e->ce_dir);
}

t_symbol *canvas_getdir(t_canvas *x)
{
    t_canvasenvironment *e = canvas_getenv(x);
    return (e->ce_dir);
}

/* -------------------  editors ------------------------------ */

static t_editor *editor_new(void)
{
    t_editor *x = (t_editor *)getbytes(sizeof(*x));
    memset((char *)x, 0, sizeof(*x));
    x->e_connectbuf = binbuf_new();
    x->e_deleted = binbuf_new();
    return (x);
}

static void editor_free(t_editor *x, t_glist *y)
{
    glist_noselect(y);
    freebytes((void *)x, sizeof(*x));
}
    
/* ------------------------ managing the selection ----------------- */

int glist_isselected(t_glist *x, t_gobj *y)
{
    if (x->gl_editor)
    {
    	t_selection *sel;
    	for (sel = x->gl_editor->e_selection; sel; sel = sel->sel_next)
    	    if (sel->sel_what == y) return (1);
    }
    return (0);
}

    /* call this for unselected objects only */
void glist_select(t_glist *x, t_gobj *y)
{
    if (x->gl_editor)
    {
	t_selection *sel = (t_selection *)getbytes(sizeof(*sel));
	    /* LATER #ifdef out the following check */
	if (glist_isselected(x, y)) bug("glist_select");
	sel->sel_next = x->gl_editor->e_selection;
	sel->sel_what = y;
	x->gl_editor->e_selection = sel;
	gobj_select(y, x, 1);
    }
}

    /* call this for selected objects only */
void glist_deselect(t_glist *x, t_gobj *y)
{
    static int reenter = 0;
    if (reenter) return;
    reenter = 1;
    if (x->gl_editor)
    {
	t_selection *sel, *sel2;
	t_rtext *z = 0;
	if (!glist_isselected(x, y)) bug("glist_deselect");
	if (x->gl_editor->e_textedfor)
	{
	    t_rtext *fuddy = glist_findrtext(x, (t_text *)y);
	    if (x->gl_editor->e_textedfor == fuddy)
	    {
	    	if (x->gl_editor->e_textdirty)
	    	{
		    z = fuddy;
    	    	    canvas_stowconnections(glist_getcanvas(x));
    	    	}
    	    	gobj_activate(y, x, 0);
	    }
	}
#if 0
	post("deselect %x", y);
	for (sel3 = x->gl_editor->e_selection; sel3; sel3 = sel3->sel_next)
	{
	     post("list %x->%x", sel3, sel3->sel_what);
	     if (count++ > 30) exit(1);
	}
#endif
	if ((sel = x->gl_editor->e_selection)->sel_what == y)
	{
     	    x->gl_editor->e_selection = x->gl_editor->e_selection->sel_next;
    	    gobj_select(sel->sel_what, x, 0);
    	    freebytes(sel, sizeof(*sel));
	}
	else
	{
    	    for (sel = x->gl_editor->e_selection; sel2 = sel->sel_next;
    	    	sel = sel2)
    	    {
    		if (sel2->sel_what == y)
    		{
    	    	    sel->sel_next = sel2->sel_next;
     	    	    gobj_select(sel2->sel_what, x, 0);
    	    	    freebytes(sel2, sizeof(*sel2));
    	    	    break;
    		}
    	    }
	}
	if (z)
    	{
    	    char *buf;
    	    int bufsize;

    	    rtext_gettext(z, &buf, &bufsize);
    	    text_setto((t_text *)y, x, buf, bufsize);
    	    x->gl_editor->e_textedfor = 0;
    	}

    }
    reenter = 0;
}

void glist_noselect(t_glist *x)
{
    if (x->gl_editor) while (x->gl_editor->e_selection)
    	glist_deselect(x, x->gl_editor->e_selection->sel_what);
}

void glist_selectall(t_glist *x)
{
    if (x->gl_editor)
    {
    	glist_noselect(x);
    	if (x->gl_list)
    	{
    	    t_selection *sel = (t_selection *)getbytes(sizeof(*sel));
    	    t_gobj *y = x->gl_list;
    	    x->gl_editor->e_selection = sel;
    	    sel->sel_what = y;
    	    gobj_select(y, x, 1);
    	    while (y = y->g_next)
    	    {
    	    	t_selection *sel2 = (t_selection *)getbytes(sizeof(*sel2));
    	    	sel->sel_next = sel2;
    	    	sel = sel2;
    	    	sel->sel_what = y;
    	    	gobj_select(y, x, 1);
    	    }
    	    sel->sel_next = 0;
    	}
    }
}

/* ---------------- other functions common to glists ----------------- */

void glist_add(t_glist *x, t_gobj *y)
{
    y->g_next = 0;
    if (!x->gl_list) x->gl_list = y;
    else
    {
    	t_gobj *y2;
    	for (y2 = x->gl_list; y2->g_next; y2 = y2->g_next);
    	y2->g_next = y;
    }
    if (glist_getcanvas(x)->x_vis) gobj_vis(y, x, 1);
}

void glist_delete(t_glist *x, t_gobj *y)
{
    t_gobj *g;
    int restoredsp = 0;
    	/*  LATER decide whether all visible glists must have an editor? */
    if (glist_getcanvas(x)->x_vis && x->gl_editor)
    {
    	if (x->gl_editor->e_grab == y) x->gl_editor->e_grab = 0;
    	if (glist_isselected(x, y)) glist_deselect(x, y);
    }
    if (canvas_dspstate && zgetfn(&y->g_pd, gensym("dsp")))
    {
    	canvas_stop_dsp();
    	restoredsp = 1;
    }
    gobj_delete(y, x);
    if (glist_getcanvas(x)->x_vis) gobj_vis(y, x, 0);
    if (x->gl_list == y) x->gl_list = y->g_next;
    else for (g = x->gl_list; g; g = g->g_next)
    	if (g->g_next == y)
    {
    	g->g_next = y->g_next;
    	break;
    }
    pd_free(&y->g_pd);
    if (restoredsp) canvas_start_dsp();
}

void glist_retext(t_glist *glist, t_text *y)
{
    t_canvas *x = glist_getcanvas(glist);
    if (x->x_vis)
    {
    	t_rtext *rt = glist_findrtext(glist, y);
    	rtext_retext(rt);
    }
}

void glist_grab(t_glist *x, t_gobj *y, t_int xpos, t_int ypos)
{
    x->gl_editor->e_onmotion = MA_PASSOUT;
    x->gl_editor->e_grab = y;
    x->gl_editor->e_xwas = xpos;
    x->gl_editor->e_ywas = ypos;
}

t_canvas *glist_getcanvas(t_glist *x)
{
    while (x->gl_owner) x = x->gl_owner;
    if (pd_class(&x->gl_pd) != canvas_class) bug("glist_getcanvas");
    return((t_canvas *)x);
}

int glist_isvisible(t_glist *x)
{
    while (x->gl_owner) x = x->gl_owner;
    if (pd_class(&x->gl_pd) != canvas_class) bug("glist_isvisible");
    return(((t_canvas *)x)->x_vis);
}

char *glist_getfont(t_glist *x)
{
    return ((glist_getcanvas((t_glist *)x)->x_font->s_name));
}

void glist_init(t_glist *x)
{
    x->gl_owner = 0;
    x->gl_editor = 0;
    x->gl_list = 0;
    x->gl_px1 = x->gl_py1 = 0;
    x->gl_px2 = x->gl_py2 = 100;
    x->gl_gobj.g_next = 0;
}

/* ---------------- generic widget behavior ------------------------- */

void gobj_getrect(t_gobj *x, t_glist *glist, int *x1, int *y1,
    int *x2, int *y2)
{
    if (x->g_pd->c_wb)
    	(*(x->g_pd)->c_wb->w_getrectfn)(x, glist, x1, y1, x2, y2);
}

void gobj_displace(t_gobj *x, t_glist *glist, int dx, int dy)
{
    if (x->g_pd->c_wb)
    	(*(x->g_pd)->c_wb->w_displacefn)(x, glist, dx, dy);
}

void gobj_select(t_gobj *x, t_glist *glist, int state)
{
    if (x->g_pd->c_wb)
    	(*(x->g_pd)->c_wb->w_selectfn)(x, glist, state);
}

void gobj_activate(t_gobj *x, t_glist *glist, int state)
{
    if (x->g_pd->c_wb)
    	(*x->g_pd->c_wb->w_activatefn)(x, glist, state);
}

void gobj_delete(t_gobj *x, t_glist *glist)
{
    if (x->g_pd->c_wb)
    	(*(x->g_pd)->c_wb->w_deletefn)(x, glist);
}

void gobj_vis(t_gobj *x, t_glist *glist, int flag)
{
    if (x->g_pd->c_wb)
    	(*(x->g_pd)->c_wb->w_visfn)(x, glist, flag);
}

void gobj_save(t_gobj *x, t_binbuf *b)
{
    if (x->g_pd->c_wb)
    	(*(x->g_pd)->c_wb->w_savefn)(x, b);
}

/* --------------- traversing the set of lines in a canvas ----------- */

typedef struct linetraverser
{
    t_canvas *tr_x;
    t_object *tr_ob;
    int tr_nout;
    int tr_outno;
    t_object *tr_ob2;
    int tr_nin;
    int tr_inno;
    int tr_x11, tr_y11, tr_x12, tr_y12;
    int tr_x21, tr_y21, tr_x22, tr_y22;
    int tr_lx1, tr_ly1, tr_lx2, tr_ly2;
    t_outconnect *tr_nextoc;
    int tr_nextoutno;
} t_linetraverser;

static void linetraverser_start(t_linetraverser *t, t_canvas *x)
{
    t->tr_ob = 0;
    t->tr_x = x;
    t->tr_nextoc = 0;
    t->tr_nextoutno = t->tr_nout = 0;
}

static t_outconnect *linetraverser_next(t_linetraverser *t)
{
    t_outconnect *rval = t->tr_nextoc;
    t_gobj *y;
    int outno;
    while (!rval)
    {
    	t_outlet *o;
    	outno = t->tr_nextoutno;
	while (outno == t->tr_nout)
	{
    	    t_gobj *y;
    	    t_object *ob = 0;
    	    if (!t->tr_ob) y = t->tr_x->x_list;
    	    else y = t->tr_ob->ob_g.g_next;
    	    for (; y; y = y->g_next)
    		if (ob = pd_checkobject(&y->g_pd)) break;
    	    if (!ob) return (0);
    	    t->tr_ob = ob;
    	    t->tr_nout = obj_noutlets(ob);
    	    outno = 0;
   	    if (t->tr_x->x_vis)
   	    	gobj_getrect(y, &t->tr_x->x_glist,
   	    	    &t->tr_x11, &t->tr_y11, &t->tr_x12, &t->tr_y12);
   	    else t->tr_x11 = t->tr_y11 = t->tr_x12 = t->tr_y12 = 0;
	}
	t->tr_nextoutno = outno + 1;
	rval = obj_starttraverseoutlet(t->tr_ob, &o, outno);
    	t->tr_outno = outno;
    }
    t->tr_nextoc = obj_nexttraverseoutlet(rval, &t->tr_ob2, &t->tr_inno);
    t->tr_nin = obj_ninlets(t->tr_ob2);
    if (!t->tr_nin) bug("drawline");
    if (t->tr_x->x_vis)
    {
    	int inplus = (t->tr_nin == 1 ? 1 : t->tr_nin - 1);
    	int outplus = (t->tr_nout == 1 ? 1 : t->tr_nout - 1);
    	gobj_getrect(&t->tr_ob2->ob_g, &t->tr_x->x_glist,
   	    &t->tr_x21, &t->tr_y21, &t->tr_x22, &t->tr_y22);
    	t->tr_lx1 = t->tr_x11 +
   	    ((t->tr_x12 - t->tr_x11 - IOWIDTH) * t->tr_outno)/outplus + IOMIDDLE;
    	t->tr_ly1 = t->tr_y12;
    	t->tr_lx2 = t->tr_x21 +
   	    ((t->tr_x22 - t->tr_x21 - IOWIDTH) * t->tr_inno)/inplus +
   	    	IOMIDDLE;
    	t->tr_ly2 = t->tr_y21;
    }
    else
    {
    	t->tr_x21 = t->tr_y21 = t->tr_x22 = t->tr_y22 = 0;
    	t->tr_lx1 = t->tr_ly1 = t->tr_lx2 = t->tr_ly2 = 0;
    }
    
    return (rval);
}

static void linetraverser_skipobject(t_linetraverser *t)
{
    t->tr_nextoc = 0;
    t->tr_nextoutno = t->tr_nout;
}

/* -------------------- the canvas object -------------------------- */

void *canvas_new(t_symbol *sel, int argc, t_atom *argv)
{
    t_canvas *x = (t_canvas *)pd_new(canvas_class);
    t_subcanvas *y = (t_subcanvas *)pd_new(subcanvas_class);
    char buf[30];
    t_canvas *super = canvas_getcurrent();
    t_symbol *s;
    int vis, width, height;
    if (argc == 2)
    {
    	if (argv->a_type == A_SYMBOL)   /* obsolete: name, vis */
    	{
    	    width = 600;
    	    height = 400;
    	    s = atom_getsymarg(0, argc, argv);
    	    vis = atom_getintarg(1, argc, argv);
    	}
    	else   /* current: width, height */
    	{
    	    width = atom_getintarg(0, argc, argv);;
    	    height = atom_getintarg(1, argc, argv);;
    	    s = &s_;
    	    vis = 0;
    	}
    }
    else if (argc == 4)     /* width, height,  */
    {
    	width = atom_getintarg(0, argc, argv);;
    	height = atom_getintarg(1, argc, argv);;
    	s = atom_getsymarg(2, argc, argv);
    	vis = atom_getintarg(3, argc, argv);
    }
    else  /* all default */
    {
    	width = 600;
    	height = 400;
    	vis = 0;
    	s = &s_;
    }
    if (canvas_newdirectory)
    {
    	t_canvasenvironment *env = x->x_env =
    	    (t_canvasenvironment *)getbytes(sizeof(*x->x_env));
    	env->ce_dir = canvas_newdirectory;
    	env->ce_argc = canvas_newargc;
    	env->ce_argv = canvas_newargv;
    	canvas_newdirectory = 0;
    	canvas_newargc = 0;
    	canvas_newargv = 0;
    }
    else x->x_env = 0;
    glist_init(&x->x_glist);
    sprintf(buf, ".x%x", x);
    pd_bind(&x->x_pd, gensym(buf));
    x->x_glist.gl_name = (*s->s_name ? s : 
    	(canvas_newfilename ? canvas_newfilename : gensym("Pd")));
    pd_bind(&x->x_pd, x->x_glist.gl_name);

    x->x_dirty = 0;
    x->x_edit = !strncmp(x->x_glist.gl_name->s_name, "Untitled", 8);
    x->x_vis = 0;
    x->x_loading = 1;
    x->x_willvis = vis;
    x->x_font = sys_defaultfont;
    pd_pushsym(&x->x_pd);

    y->x_subcanvas = x;
    y->x_supercanvas = super;
    x->x_parent = y;

    return(x);
}

extern t_pd *newest;

void canvas_reflecttitle(t_canvas *x)
{
    sys_vgui("wm title .x%x {%s%c - %s}\n", 
    	x, x->x_glist.gl_name->s_name, (x->x_dirty? '*' : ' '),
    	    canvas_getdir(x)->s_name);
}

void canvas_dirty(t_canvas *x, t_int n)
{
    t_canvas *x2 = canvas_getrootfor(x);
    if (n != x2->x_dirty)
    {
    	x2->x_dirty = n;
    	canvas_reflecttitle(x);
    }
}

void canvas_vis(t_canvas *x, int flag)
{
    t_gobj *y;
    if (flag == x->x_vis) return;
    if (flag)
    {
    	x->x_glist.gl_editor = editor_new();
    	sys_vgui("pdtk_canvas_new .x%x\n", x);
    	canvas_reflecttitle(x);
    	for (y = x->x_list; y; y = y->g_next) gobj_vis(y, &x->x_glist, 1);
    	x->x_vis = flag;
    	canvas_drawlines(x);
    	    /* add to Pd window find menu */
    	sys_vgui(
    	    ".mbar.find.menu add command -label %s -command {raise .x%x}\n",
    	    	x->x_glist.gl_name->s_name, x);
    	    	/* add to the window list. */
    	x->x_next = canvas_list;
    	canvas_list = x;
    }
    else
    {
    	int i;
    	t_canvas *x2;
    	for (y = x->x_list; y; y = y->g_next) gobj_vis(y, &x->x_glist, 0);
    	editor_free(x->x_glist.gl_editor, &x->x_glist);
    	x->x_glist.gl_editor = 0;
   	sys_vgui("destroy .x%x\n", x);
    	for (i = 1, x2 = x; x2; x2 = x2->x_next, i++);
    	sys_vgui(".mbar.find.menu delete %d\n", i);
    	
    	    /* take it off the window list */
	if (x == canvas_list) canvas_list = x->x_next;
	else
	{
    	    t_canvas *y;
    	    for (y = canvas_list; y->x_next != x; y = y->x_next);
    	    y->x_next = x->x_next;
	}
    	x->x_vis = flag;
    }
}

void canvas_free(t_canvas *x)
{
    t_gobj *y;
    char buf[30];
    int restoredsp = 0;

    if (canvas_dspstate)
    {
    	canvas_stop_dsp();
    	restoredsp = 1;
    }
    if (canvas_editing == x) canvas_editing = 0;
    sprintf(buf, ".x%x", x);
    pd_unbind(&x->x_pd, gensym(buf));
    pd_unbind(&x->x_pd, x->x_glist.gl_name);
    while (y = x->x_glist.gl_list) glist_delete(&x->x_glist, y);
    canvas_vis(x, 0);
    if (restoredsp) canvas_start_dsp();
}

/* ------------------------ event handling ------------------------ */

static void canvas_setcursor(t_canvas *x, char *name)
{
    static t_canvas *xwas;
    static char *namewas;
    if (xwas != x || namewas != name)
    {
    	sys_vgui(".x%x configure -cursor %s\n", x, name);
    	xwas = x;
    	namewas = name;
    }
}

#define NOMOD 0
#define SHIFTMOD 1
#define CTRLMOD 2
#define ALTMOD 3
#define DBLMOD 4

t_gobj *canvas_hitbox(t_canvas *x, int xpos, int ypos,
    int *x1p, int *y1p, int *x2p, int *y2p)
{
    t_gobj *y;
    for (y = x->x_list; y; y = y->g_next)
    {
    	int x1, y1, x2, y2;
    	gobj_getrect(y, &x->x_glist, &x1, &y1, &x2, &y2);
    	if (xpos >= x1 && xpos <= x2 && ypos >= y1 && ypos <= y2)
    	{
    	    *x1p = x1;
    	    *y1p = y1;
    	    *x2p = x2;
    	    *y2p = y2;
    	    return (y);
    	}
    }
    return (0);
}

    /* single mouse click */
void canvas_doclick(t_canvas *x, t_int xpos, t_int ypos, t_int which,
    t_int mod, t_int doit)
{
    t_gobj *y;
    
    if (!x->x_glist.gl_editor)
    {
    	bug("editor");
    	return;
    }
    if ((!x->x_edit) && (!mod || mod == DBLMOD)) mod = CTRLMOD; 
    x->x_glist.gl_editor->e_lastmoved = 0;
    if (doit) x->x_glist.gl_editor->e_grab = 0;
    /* post("click %d %d %d %d", xpos, ypos, which, mod); */
    
    if (x->x_glist.gl_editor->e_onmotion == MA_NONE)
    {
    	int x1, y1, x2, y2;
    	x->x_glist.gl_editor->e_xwas = xpos;
    	x->x_glist.gl_editor->e_ywas = ypos;
    	
    	if (doit && x->x_glist.gl_editor->e_textedfor)
    	    glist_noselect(&x->x_glist);
        if (y = canvas_hitbox(x, xpos, ypos, &x1, &y1, &x2, &y2))
    	{
    	    	/* hit a rectangle. */
    	    if (mod == NOMOD)
    	    {
    	    	    /* look for an outlet */
    	    	t_object *ob = pd_checkobject(&y->g_pd);
    	    	int noutlet;
    	    	if (ob && (noutlet = obj_noutlets(ob)) && ypos >= y2-3)
    	    	{
    	    	    int width = x2 - x1;
    	    	    int nout1 = (noutlet > 1 ? noutlet - 1 : 1);
    	    	    int closest = ((xpos-x1) * (nout1) + width/2)/width;
    	    	    int hotspot = x1 +
    	    		(width - IOWIDTH) * closest / (nout1);
    	    	    if (closest < noutlet &&
    	    	    	xpos >= hotspot && xpos < hotspot + IOWIDTH)
    	    	    {
    	    	    	if (doit)
    	    	    	{
    	    	    	    x->x_glist.gl_editor->e_onmotion = MA_CONNECT;
    	    	    	    x->x_glist.gl_editor->e_xwas = xpos;
    	    	    	    x->x_glist.gl_editor->e_ywas = ypos;
    	    	    	    sys_vgui(
    	    	    	    	".x%x.c create line %d %d %d %d -tags x\n",
    	    	    	    	    x, xpos, ypos, xpos, ypos);
    	    	    	}    	    	    	    	
    	    	    	else canvas_setcursor(x, "circle");
    	    	    	return;
    	    	    }
    	    	}
    	    	    /* not in an outlet; select and move */
    	    	if (doit)
    	    	{
    	    	    if (!glist_isselected(&x->x_glist, y))
    	    	    {
    	    		glist_noselect(&x->x_glist);
    	    		glist_select(&x->x_glist, y);
    	    	    }
    	    	    x->x_glist.gl_editor->e_onmotion = MA_MOVE;
    	    	}
    	    	else canvas_setcursor(x, "hand2"); 
    	    }
    	    else if (mod == SHIFTMOD)
    	    {
    	    	if (doit)
    	    	{
    	    	    if (glist_isselected(&x->x_glist, y))
    	    	    	glist_deselect(&x->x_glist, y);
    	    	    else glist_select(&x->x_glist, y);
    	    	}
    	    }
    	    else if (mod == CTRLMOD)
    	    {
    	    	if (doit && zgetfn(&y->g_pd, gensym("click")))
    	    	    vmess(&y->g_pd, gensym("click"), "iii", xpos, ypos, 1);
    	    	else canvas_setcursor(x, "left_ptr");
    	    }
    	    /* if (doit) canvas_setcursor(x, "left_ptr"); */
    	    return;
    	}
    	if (mod == CTRLMOD)
    	{
    	    canvas_setcursor(x, "left_ptr");
    	    return;
    	}
    	    /* having failed to find a box, we try lines now. */
	if (!mod || mod == DBLMOD)
	{
    	    t_linetraverser t;
    	    t_outconnect *oc;
    	    float fx = xpos, fy = ypos;
    	    linetraverser_start(&t, glist_getcanvas(&x->x_glist));
    	    while (oc = linetraverser_next(&t))
    	    {
    		float lx1 = t.tr_lx1, ly1 = t.tr_ly1,
    	    	    lx2 = t.tr_lx2, ly2 = t.tr_ly2;
    		float area = (lx2 - lx1) * (fy - ly1) -
    		    (ly2 - ly1) * (fx - lx1);
    		float dsquare = (lx2-lx1) * (lx2-lx1) + (ly2-ly1) * (ly2-ly1);
    		if (area * area >= 50 * dsquare) continue;
    		if ((lx2-lx1) * (fx-lx1) + (ly2-ly1) * (fy-ly1) < 0) continue;
    		if ((lx2-lx1) * (lx2-fx) + (ly2-ly1) * (ly2-fy) < 0) continue;
    		if (doit)
    		{
    	    	    sys_vgui(".x%x.c delete l%x\n",
    	    	    	glist_getcanvas(&x->x_glist), oc);
    	    	    obj_disconnect(t.tr_ob, t.tr_outno, t.tr_ob2, t.tr_inno);
    		}
    		else canvas_setcursor(x, "X_cursor");
    		return;
    	    }
	}
	canvas_setcursor(x, "hand2");
	if (doit)
	{
	    if (mod != SHIFTMOD) glist_noselect(&x->x_glist);
    	    sys_vgui(".x%x.c create rectangle %d %d %d %d -tags x\n",
    	    	  x, xpos, ypos, xpos, ypos);
    	    x->x_glist.gl_editor->e_onmotion = MA_REGION;
    	}
    }
    
}

void canvas_click(t_canvas *x, t_int xpos, t_int ypos, t_int which, t_int mod)
{
    canvas_doclick(x, xpos, ypos, which, mod, 1);
}

void canvas_doconnect(t_canvas *x, int xpos, int ypos, int which, int doit)
{
    int x11, y11, x12, y12;
    t_gobj *y1;
    int x21, y21, x22, y22;
    t_gobj *y2;
    int xwas = x->x_glist.gl_editor->e_xwas,
    	ywas = x->x_glist.gl_editor->e_ywas;
    if (doit) sys_vgui(".x%x.c delete x\n", x);
    else sys_vgui(".x%x.c coords x %d %d %d %d\n",
    	    x, x->x_glist.gl_editor->e_xwas,
    	    	x->x_glist.gl_editor->e_ywas, xpos, ypos);

    if ((y1 = canvas_hitbox(x, xwas, ywas, &x11, &y11, &x12, &y12))
    	&& (y2 = canvas_hitbox(x, xpos, ypos, &x21, &y21, &x22, &y22)))
    {
    	t_object *ob1 = pd_checkobject(&y1->g_pd);
    	t_object *ob2 = pd_checkobject(&y2->g_pd);
    	int noutlet1, ninlet2;
    	if (ob1 && ob2 && ob1 != ob2 &&
    	    (noutlet1 = obj_noutlets(ob1)) && ywas >= y12-3
    	    && (ninlet2 = obj_ninlets(ob2)))
    	{
    	    if (doit)
    	    {
    		int width1 = x12 - x11, closest1, hotspot1;
    		int width2 = x22 - x21, closest2, hotspot2;
    		if (noutlet1 > 1)
    		{
    		    closest1 = ((xwas-x11) * (noutlet1-1) + width1/2)/width1;
    		    hotspot1 = x11 +
    	    	    	(width1 - IOWIDTH) * closest1 / (noutlet1-1);
    	    	}
    	    	else closest1 = 0, hotspot1 = x11;
    	    	
    		if (ninlet2 > 1)
    		{
    		    closest2 = ((xpos-x21) * (ninlet2-1) + width2/2)/width2;
    		    hotspot2 = x21 +
    	    	    	(width2 - IOWIDTH) * closest2 / (ninlet2-1);
    	    	}
    	    	else closest2 = 0, hotspot2 = x21;
    	    	
    		if (closest1 < noutlet1 &&
    	    	    xwas >= hotspot1 && xwas < hotspot1 + IOWIDTH &&
    	    	    closest2 < ninlet2)
    		{
    	    	    t_outconnect *oc =
    	    		obj_connect(ob1, closest1, ob2, closest2);
    	    	    int lx1 = x11 +
   	    		(noutlet1 > 1 ?
   	    		    ((x12-x11-IOWIDTH) * closest1)/(noutlet1-1) : 0)
   	    		    	 + IOMIDDLE;
    	    	    int ly1 = y12;
    	    	    int lx2 = x21 +
    	    	    	(ninlet2 > 1 ?
   	    		    ((x22-x21-IOWIDTH) * closest2)/(ninlet2-1) : 0)
   	    		    	+ IOMIDDLE;
    	    	    int ly2 = y21;

    	    	    sys_vgui(".x%x.c create line %d %d %d %d -tags l%x\n",
			glist_getcanvas(&x->x_glist),
		    	    lx1, ly1, lx2, ly2, oc);
    	    	}
    	    	else bug("canvas_doconnect");
    	    }
    	    else canvas_setcursor(x, "circle");
    	    return;
    	}
    }
    canvas_setcursor(x, "hand2");
}

void canvas_doregion(t_canvas *x, int xpos, int ypos, int doit)
{
    if (doit)
    {
    	t_gobj *y;
    	int lox, loy, hix, hiy;
    	if (x->x_glist.gl_editor->e_xwas < xpos)
    	    lox = x->x_glist.gl_editor->e_xwas, hix = xpos;
    	else hix = x->x_glist.gl_editor->e_xwas, lox = xpos;
    	if (x->x_glist.gl_editor->e_ywas < ypos)
    	    loy = x->x_glist.gl_editor->e_ywas, hiy = ypos;
    	else hiy = x->x_glist.gl_editor->e_ywas, loy = ypos;
    	for (y = x->x_list; y; y = y->g_next)
    	{
    	    int x1, y1, x2, y2;
    	    gobj_getrect(y, &x->x_glist, &x1, &y1, &x2, &y2);
    	    if (hix >= x1 && lox <= x2 && hiy >= y1 && loy <= y2
    	    	&& !glist_isselected(&x->x_glist, y))
    	    	    glist_select(&x->x_glist, y);
    	}
    	sys_vgui(".x%x.c delete x\n", x);
    	x->x_glist.gl_editor->e_onmotion = 0;
    }
    else sys_vgui(".x%x.c coords x %d %d %d %d\n",
    	    x, x->x_glist.gl_editor->e_xwas,
    	    	x->x_glist.gl_editor->e_ywas, xpos, ypos);
}

void canvas_mouseup(t_canvas *x, t_int xpos, t_int ypos, t_int which,
    t_int width, t_int height)
{
    /* post("mouseup %d %d %d", xpos, ypos, which); */
    if (!x->x_glist.gl_editor)
    {
    	bug("editor");
    	return;
    }
    if (x->x_glist.gl_editor->e_onmotion == MA_CONNECT)
    	canvas_doconnect(x, xpos, ypos, which, 1);
    else if (x->x_glist.gl_editor->e_onmotion == MA_REGION)
    	canvas_doregion(x, xpos, ypos, 1);
    else if (x->x_glist.gl_editor->e_onmotion == MA_MOVE)
    {
    	    /* after motion, if there's only one item selected, activate it */
    	if (x->x_glist.gl_editor->e_selection &&
    	    !(x->x_glist.gl_editor->e_selection->sel_next))
    	    	gobj_activate(x->x_glist.gl_editor->e_selection->sel_what,
    	    	    &x->x_glist, 1);
    }
    x->x_glist.gl_editor->e_onmotion = MA_NONE;
    if (width) x->x_glist.gl_px2 = width;
    if (height) x->x_glist.gl_py2 = height;
}

void canvas_key(t_canvas *x, int n)
{
    /* post("key %d %c", n, n); */
    if (n == '\r') n = '\n';
    if (!x->x_glist.gl_editor)
    {
    	bug("editor");
    	return;
    }
    if (x->x_glist.gl_editor->e_grab)
    	vmess(&x->x_glist.gl_editor->e_grab->g_pd, gensym("key"), "i", n);
    else if (x->x_glist.gl_editor->e_textedfor)
    {
    	rtext_key(x->x_glist.gl_editor->e_textedfor, n);
    	x->x_glist.gl_editor->e_textdirty = 1;
    	canvas_dirty(x, 1);
    }
}

void canvas_motion(t_canvas *x, t_int xpos, t_int ypos)
{ 
    /* post("motion %d %d", xpos, ypos); */
    if (!x->x_glist.gl_editor)
    {
    	bug("editor");
    	return;
    }
    if (x->x_glist.gl_editor->e_onmotion == MA_MOVE)
    {
    	t_selection *y;
    	for (y = x->x_glist.gl_editor->e_selection; y; y = y->sel_next)
    	    gobj_displace(y->sel_what, &x->x_glist, xpos - x->x_glist.gl_editor->e_xwas,
    	    	ypos - x->x_glist.gl_editor->e_ywas);
    	x->x_glist.gl_editor->e_xwas = xpos;
    	x->x_glist.gl_editor->e_ywas = ypos;
    	canvas_dirty(x, 1);
    }
    else if (x->x_glist.gl_editor->e_onmotion == MA_REGION)
    	canvas_doregion(x, xpos, ypos, 0);
    else if (x->x_glist.gl_editor->e_onmotion == MA_CONNECT)
    	canvas_doconnect(x, xpos, ypos, 0, 0);
    else if (x->x_glist.gl_editor->e_onmotion == MA_PASSOUT)
    {
    	vmess(&x->x_glist.gl_editor->e_grab->g_pd, gensym("motion"), "ii",
    	    xpos - x->x_glist.gl_editor->e_xwas,
    	    ypos - x->x_glist.gl_editor->e_ywas);
    	x->x_glist.gl_editor->e_xwas = xpos;
    	x->x_glist.gl_editor->e_ywas = ypos;
    }
    else canvas_doclick(x, xpos, ypos, 0, 0, 0);
    
    x->x_glist.gl_editor->e_lastmoved = 1;
}

/* ----------------- lines ---------- */

static void canvas_drawlines(t_canvas *x)
{
    t_linetraverser t;
    t_outconnect *oc;
    {
    	linetraverser_start(&t, x);
    	while (oc = linetraverser_next(&t))
    	    sys_vgui(".x%x.c create line %d %d %d %d -tags l%x\n",
		    glist_getcanvas(&x->x_glist),
		    	t.tr_lx1, t.tr_ly1, t.tr_lx2, t.tr_ly2, oc);
    }
}

void canvas_fixlinesfor(t_canvas *x, t_text *text)
{
    t_linetraverser t;
    t_outconnect *oc;

    linetraverser_start(&t, x);
    while (oc = linetraverser_next(&t))
    {
    	if (t.tr_ob == text || t.tr_ob2 == text)
    	{
    	    sys_vgui(".x%x.c coords l%x %d %d %d %d\n",
		glist_getcanvas(&x->x_glist), oc,
		    t.tr_lx1, t.tr_ly1, t.tr_lx2, t.tr_ly2);
    	}
    }
}

    /* kill all lines for the object */
void canvas_deletelinesfor(t_canvas *x, t_text *text)
{
    if (x->x_glist.gl_editor)
    {
    	t_linetraverser t;
    	t_outconnect *oc;
    	linetraverser_start(&t, x);
    	while (oc = linetraverser_next(&t))
    	{
    	    if (t.tr_ob == text || t.tr_ob2 == text)
    	    {
    	    	sys_vgui(".x%x.c delete l%x\n",
    	    	    glist_getcanvas(&x->x_glist), oc);
    	    	obj_disconnect(t.tr_ob, t.tr_outno, t.tr_ob2, t.tr_inno);
    	    }
    	}
    }
}

/* ----------------------------- window stuff ----------------------- */

void canvas_print(t_canvas *x, t_symbol *s)
{
    if (*s->s_name) sys_vgui(".x%x.c postscript -file %s\n", x, s->s_name);
    else sys_vgui(".x%x.c postscript -file x.ps\n", x);
}

void canvas_tobinbuf(t_canvas *x, t_binbuf *b)
{
    t_gobj *y;
    t_linetraverser t;
    t_outconnect *oc;
    if (x->x_parent->x_supercanvas)
    	binbuf_addv(b, "ssiisi;", gensym("#N"), gensym("canvas"),
    	    (t_int)x->x_glist.gl_px2, (t_int)x->x_glist.gl_py2,
    	    	x->x_glist.gl_name, x->x_vis);
    else binbuf_addv(b, "ssii;", gensym("#N"), gensym("canvas"),
    	(t_int)x->x_glist.gl_px2, (t_int)x->x_glist.gl_py2);
    for (y = x->x_list; y; y = y->g_next) gobj_save(y, b);
    linetraverser_start(&t, x);
    while (oc = linetraverser_next(&t))
    {
    	int srcno, sinkno;
    	for (srcno = 0, y = x->x_list; y && y != &t.tr_ob->ob_g; y = y->g_next)
    	    srcno++;
    	for (sinkno = 0, y = x->x_list; y && y != &t.tr_ob2->ob_g; y = y->g_next)
    	    sinkno++;
    	binbuf_addv(b, "ssiiii;", gensym("#X"), gensym("connect"),
    	    srcno, t.tr_outno, sinkno, t.tr_inno);
    }
}

void canvas_saveto(t_canvas *x, t_symbol *filename, t_symbol *dir)
{
    t_binbuf *b = binbuf_new();
    /* post("canvas_saveto %s %s", filename->s_name, dir->s_name); */
    canvas_tobinbuf(x, b);
    if (binbuf_write(b, filename->s_name, dir->s_name)) sys_ouch();
    else
    {
    	t_canvasenvironment *e = canvas_getenv(x);
    	pd_unbind(&x->x_pd, x->x_glist.gl_name);
    	x->x_glist.gl_name = filename;
    	pd_bind(&x->x_pd, x->x_glist.gl_name);
    	e->ce_dir = dir;
    	canvas_reflecttitle(x);
    	post("saved to: %s/%s", dir->s_name, filename->s_name);
    	canvas_dirty(x, 0);
    }
}

void canvas_menusaveas(t_canvas *x)
{
    t_canvas *x2 = canvas_getrootfor(x);
    sys_vgui("pdtk_canvas_saveas .x%x %s %s\n", x2,
    	x2->x_glist.gl_name->s_name, canvas_getenv(x2)->ce_dir->s_name);
}

void canvas_menuclose(t_canvas *x, t_int force)
{
    if (x->x_parent->x_supercanvas) canvas_vis(x, 0);
    else if (force || (!x->x_dirty)) pd_free(&x->x_pd);
    else sys_vgui("pdtk_check {This window has been modified.  Close anyway?}\
     {.x%x menuclose 1\\\n}\n", x);
}

void canvas_stowconnections(t_canvas *x)
{
    t_gobj *selhead = 0, *seltail = 0, *nonhead = 0, *nontail = 0, *y, *y2;
    t_linetraverser t;
    t_outconnect *oc;
    if (!x->x_glist.gl_editor) return;
    	/* split list to "selected" and "unselected" parts */ 
    for (y = x->x_glist.gl_list; y; y = y2)
    {
    	y2 = y->g_next;
    	if (glist_isselected(&x->x_glist, y))
    	{
    	    if (seltail)
    	    {
    	    	seltail->g_next = y;
    	    	seltail = y;
    	    	y->g_next = 0;
    	    }
    	    else
    	    {
    	    	selhead = seltail = y;
    	    	seltail->g_next = 0;
    	    }
    	}
    	else
    	{
    	    if (nontail)
    	    {
    	    	nontail->g_next = y;
    	    	nontail = y;
    	    	y->g_next = 0;
    	    }
    	    else
    	    {
    	    	nonhead = nontail = y;
    	    	nontail->g_next = 0;
    	    }
    	}
    }
    	/* move the selected part to the end */
    if (!nonhead) x->x_glist.gl_list = selhead;
    else x->x_glist.gl_list = nonhead, nontail->g_next = selhead;

    	/* add connections to binbuf */
    binbuf_clear(x->x_glist.gl_editor->e_connectbuf);
    linetraverser_start(&t, x);
    while (oc = linetraverser_next(&t))
    {
    	int srcno = 0, sinkno = 0;
    	int s1 = glist_isselected(&x->x_glist, &t.tr_ob->ob_g);
    	int s2 = glist_isselected(&x->x_glist, &t.tr_ob2->ob_g);
    	if (s1 != s2)
    	{
    	    for (y = x->x_list; y && y != &t.tr_ob->ob_g; y = y->g_next)
    	    	srcno++;
    	    for (y = x->x_list; y && y != &t.tr_ob2->ob_g; y = y->g_next)
    	    	sinkno++;
    	    binbuf_addv(x->x_glist.gl_editor->e_connectbuf, "ssiiii;",
    	    	gensym("#X"), gensym("connect"),
    	    	    srcno, t.tr_outno, sinkno, t.tr_inno);
    	}
    }
}

void canvas_restoreconnections(t_canvas *x)
{
    pd_bind(&x->x_pd, gensym("#X"));
    binbuf_eval(x->x_glist.gl_editor->e_connectbuf, 0, 0, 0);
    pd_unbind(&x->x_pd, gensym("#X"));
}

static t_binbuf *copy_binbuf;

static void canvas_copy(t_canvas *x)
{
    t_gobj *y;
    t_linetraverser t;
    t_outconnect *oc;
    binbuf_clear(copy_binbuf);
    for (y = x->x_glist.gl_list; y; y = y->g_next)
    	if (glist_isselected(&x->x_glist, y))
    	    gobj_save(y, copy_binbuf);
    linetraverser_start(&t, x);
    while (oc = linetraverser_next(&t))
    {
    	int srcno = 0, sinkno = 0;
    	if (glist_isselected(&x->x_glist, &t.tr_ob->ob_g)
    	    && glist_isselected(&x->x_glist, &t.tr_ob2->ob_g))
    	{
    	    for (y = x->x_list; y && y != &t.tr_ob->ob_g; y = y->g_next)
    	    	if (glist_isselected(&x->x_glist, y)) srcno++;
    	    for (y = x->x_list; y && y != &t.tr_ob2->ob_g; y = y->g_next)
    	    	if (glist_isselected(&x->x_glist, y)) sinkno++;
    	    binbuf_addv(copy_binbuf, "ssiiii;", gensym("#X"),
    	    	gensym("connect"), srcno, t.tr_outno, sinkno, t.tr_inno);
    	}
    }
}

static void canvas_cut(t_canvas *x)
{
    t_gobj *y, *y2;
    int restoredsp;
    canvas_copy(x);
    if (restoredsp = canvas_dspstate) canvas_stop_dsp();
    while (1)	/* this is pretty wierd...  should rewrite it */
    {
    	for (y = x->x_glist.gl_list; y; y = y2)
    	{
    	    y2 = y->g_next;
    	    if (glist_isselected(&x->x_glist, y))
    	    {
    		glist_delete(&x->x_glist, y);
#if 0
    		if (y2) post("cut 5 %x %x", y2, y2->g_next);
    		else post("cut 6");
#endif
    		goto next;
    	    }
    	}
    	goto restore;
    next: ;
    }
restore:
    if (restoredsp) canvas_start_dsp();
    canvas_dirty(x, 1);
}

static void canvas_paste(t_canvas *x)
{
    t_gobj *prev, *last;
    int restoredsp;
    if (restoredsp = canvas_dspstate) canvas_stop_dsp();
    glist_noselect(&x->x_glist);
    prev = x->x_glist.gl_list;
    x->x_glist.gl_list = 0;
    pd_bind(&x->x_pd, gensym("#X"));
    binbuf_eval(copy_binbuf, 0, 0, 0);
    pd_unbind(&x->x_pd, gensym("#X"));
    glist_selectall(&x->x_glist);
    if (!x->x_glist.gl_list) x->x_glist.gl_list = prev;
    else
    {
    	for (last = x->x_glist.gl_list; last->g_next; last = last->g_next);
    	last->g_next = prev;
    }
    if (restoredsp) canvas_start_dsp();
    canvas_dirty(x, 1);
}

static void canvas_duplicate(t_canvas *x)
{
    if (x->x_glist.gl_editor->e_onmotion == MA_NONE)
    {
    	t_selection *y;
    	canvas_copy(x);
     	canvas_paste(x);
    	for (y = x->x_glist.gl_editor->e_selection; y; y = y->sel_next)
    	    gobj_displace(y->sel_what, &x->x_glist,
    	    	10, 10);
    	canvas_dirty(x, 1);
    }
}

static void canvas_texteditor(t_canvas *x)
{
    t_rtext *foo;
    char *buf;
    int bufsize;
    if (foo = x->x_glist.gl_editor->e_textedfor)
    	rtext_gettext(foo, &buf, &bufsize);
    else buf = "", bufsize = 0;
    sys_vgui("pdtk_pd_texteditor {%.*s}\n", bufsize, buf);
    
}

void glob_key(void *dummy, t_int key)
{
    if (canvas_editing) canvas_key(canvas_editing, key);
}

static void canvas_editmode(t_canvas *x, t_int yesplease)
{
    if (yesplease && x->x_edit) return;
    if (x->x_edit = !x->x_edit) canvas_setcursor(x, "hand2");
    else canvas_setcursor(x, "left_ptr");
    sys_vgui("pdtk_canvas_editval .x%x %d\n",
    	glist_getcanvas(&x->x_glist), x->x_edit);
    if (yesplease) canvas_dirty(x, 1);
}

static void canvas_menusave(t_canvas *x)
{
    t_canvas *x2 = canvas_getrootfor(x);
    if (*x2->x_glist.gl_name->s_name &&
    	strncmp(x2->x_glist.gl_name->s_name, "Untitled", 8))
    	    canvas_saveto(x2, x2->x_glist.gl_name, canvas_getenv(x2)->ce_dir);
    else canvas_menusaveas(x2);
}

static void canvas_font(t_canvas *x, t_symbol *s)
{
    x->x_font = s;
}

static void canvas_pop(t_canvas *x, t_int vis)
{
    if (vis) canvas_vis(x, 1);
    pd_popsym(&x->x_pd);
    x->x_loading = 0;
}

static void canvas_loadbang(t_canvas *x)
{
    t_gobj *y;
    t_symbol *s = gensym("loadbang");
    for (y = x->x_list; y; y = y->g_next)
    	if (pd_class(&y->g_pd) == canvas_class) canvas_loadbang((t_canvas *)y);
    for (y = x->x_list; y; y = y->g_next)
    	if ((pd_class(&y->g_pd) != canvas_class) &&
    	    zgetfn(&y->g_pd, s))
    	    	vmess(&y->g_pd, s, "");
}

void canvas_popabstraction(t_canvas *x)
{
    newest = &x->x_parent->x_ob.ob_pd;
    pd_popsym(&x->x_pd);
    x->x_loading = 0;
}


static void canvas_connect(t_canvas *x, t_intarg whoout, t_intarg outno,
    t_intarg whoin, t_intarg inno)
{
    t_gobj *src, *sink;
    t_object *objsrc, *objsink;
    t_outconnect *oc;
    int nin = whoin, nout = whoout;
    for (src = x->x_list; whoout; src = src->g_next, whoout--)
    	if (!src) goto bad;
    for (sink = x->x_list; whoin; sink = sink->g_next, whoin--)
    	if (!sink) goto bad;
    if (!(objsrc = pd_checkobject(&src->g_pd)) ||
    	!(objsink = pd_checkobject(&sink->g_pd)))
    	    goto bad;
    if (!(oc = obj_connect(objsrc, outno, objsink, inno))) goto bad;
    if (x->x_vis)
    {
    	sys_vgui(".x%x.c create line %d %d %d %d -tags l%x\n",
	    glist_getcanvas(&x->x_glist), 0, 0, 0, 0, oc);
	canvas_fixlinesfor(x, objsrc);
    }
    return;

bad:
    post("%d %d %d %d; connection failed", nout, outno, nin, inno);
}

void canvas_logerror(t_object *y)
{
#ifdef LATER
    canvas_vis(x, 1);
    if (!glist_isselected(&x->x_glist, &y->ob_g))
    	glist_select(&x->x_glist, &y->ob_g);
#endif
}

/* -------------------------- subcanvases ---------------------- */

extern void glist_text(t_glist *x, t_symbol *s, int argc, t_atom *argv);
extern void canvas_obj(t_glist *gl, t_symbol *s, int argc, t_atom *argv);
extern void canvas_msg(t_glist *gl, t_symbol *s, int argc, t_atom *argv);
extern void canvas_intatom(t_glist *gl, t_symbol *s, int argc, t_atom *argv);
extern void canvas_floatatom(t_glist *gl, t_symbol *s, int argc, t_atom *argv);
extern void canvas_symbolatom(t_glist *gl, t_symbol *s, int argc, t_atom *argv);
extern void glist_graph(t_glist *g, t_symbol *s, int argc, t_atom *argv);
extern void canvas_array(t_glist *canvas, t_symbol *s, t_intarg n,
    t_symbol *template);

static void *subcanvas_new(t_symbol *s)
{
    t_atom a[4];
    t_canvas *x, *z = canvas_getcurrent();
    t_subcanvas *y;
    if (!*s->s_name) s = gensym("/SUBPATCH/");
    SETINT(a, 600);
    SETINT(a+1, 400);
    SETSYM(a+2, s);
    SETINT(a+3, 1);
    x = canvas_new(0, 4, a);
    y = x->x_parent;
    y->x_supercanvas = z;
    canvas_pop(x, 1);
    return (y);
}

    /* this seems odd -- why can't we go straight to the canvas? */
static void subcanvas_saveto(t_subcanvas *x, t_binbuf *b)
{
    canvas_tobinbuf(x->x_subcanvas, b);
}

static void subcanvas_click(t_subcanvas *x, int xpos, int ypos, int down)
{
    if (x->x_subcanvas->x_vis)
    	sys_vgui("raise .x%x\n", glist_getcanvas(&x->x_subcanvas->x_glist));
    else canvas_vis(x->x_subcanvas, 1);
}

static void subcanvas_free(t_subcanvas *x)
{
    x->x_supercanvas = 0;
    pd_free(&x->x_subcanvas->x_pd);
}

void canvas_objfor(t_glist *gl, t_text *x, int argc, t_atom *argv);

void canvas_restore(t_canvas *x, t_symbol *s, int argc, t_atom *argv)
{
    t_subcanvas *y = x->x_parent;
    t_pd *z;
    canvas_pop(x, x->x_willvis);

    if (!(z = gensym("#X")->s_thing)) error("canvas_restore: out of context");
    else if (*z != canvas_class) error("canvas_restore: wasn't a canvas");
    else
    {
    	t_canvas *x2 = (t_canvas *)z;
    	y->x_supercanvas = x2;
    	canvas_objfor(&x2->x_glist, &y->x_ob, argc, argv);
    }
}

t_inlet *canvas_addinlet(t_canvas *x, t_pd *who)
{
    t_inlet *ip = inlet_new(&x->x_parent->x_ob, who, 0, 0);
    if (!x->x_loading &&
    	x->x_parent->x_supercanvas && x->x_parent->x_supercanvas->x_vis)
    {
    	gobj_vis(&x->x_parent->x_ob.ob_g,
    	    &x->x_parent->x_supercanvas->x_glist, 0);
    	gobj_vis(&x->x_parent->x_ob.ob_g,
    	    &x->x_parent->x_supercanvas->x_glist, 1);
    }
    return (ip);
}

void canvas_rminlet(t_canvas *x, t_inlet *ip)
{
    inlet_free(ip);
    if (x->x_parent->x_supercanvas && x->x_parent->x_supercanvas->x_vis)
    {
    	gobj_vis(&x->x_parent->x_ob.ob_g,
    	    &x->x_parent->x_supercanvas->x_glist, 0);
    	gobj_vis(&x->x_parent->x_ob.ob_g,
    	    &x->x_parent->x_supercanvas->x_glist, 1);
    }
}

t_outlet *canvas_addoutlet(t_canvas *x, t_pd *who)
{
    t_outlet *op = outlet_new(&x->x_parent->x_ob, 0);
    if (!x->x_loading &&
    	x->x_parent->x_supercanvas && x->x_parent->x_supercanvas->x_vis)
    {
    	gobj_vis(&x->x_parent->x_ob.ob_g,
    	    &x->x_parent->x_supercanvas->x_glist, 0);
    	gobj_vis(&x->x_parent->x_ob.ob_g,
    	    &x->x_parent->x_supercanvas->x_glist, 1);
    }
    return (op);
}

void canvas_rmoutlet(t_canvas *x, t_outlet *op)
{
    outlet_free(op);
    if (x->x_parent->x_supercanvas && x->x_parent->x_supercanvas->x_vis)
    {
    	gobj_vis(&x->x_parent->x_ob.ob_g,
    	    &x->x_parent->x_supercanvas->x_glist, 0);
    	gobj_vis(&x->x_parent->x_ob.ob_g,
    	    &x->x_parent->x_supercanvas->x_glist, 1);
    }
}

static int subcanvas_isabstraction(t_subcanvas *x)
{
    return (x->x_subcanvas->x_env != 0);
}

int text_isabstraction(t_text *x)
{
    return ((pd_class(&x->te_pd) == subcanvas_class) && 
    	subcanvas_isabstraction((t_subcanvas *)x));
}


t_subcanvas *canvas_getparent(t_canvas *x)
{
    return (x->x_parent);
}

    /* get the document containing this canvas */
t_canvas *canvas_getrootfor(t_canvas *x)
{
    if ((!x->x_parent) || subcanvas_isabstraction(x->x_parent)) return (x);
    else return (canvas_getrootfor(x->x_parent->x_supercanvas));
}

/* ------------------------- DSP chain handling ------------------------- */

void ugen_start_graph( void);
void ugen_add(t_object *x);
void ugen_connect(t_object *x1, int outno, t_object *x2, int inno);
void ugen_done_graph(void);
void ugen_start(void);
void ugen_stop(void);

int obj_issignaloutlet(t_object *x, int outno);

void canvas_start_dsp(void)
{
    t_symbol *dspsym = gensym("dsp");
    t_canvas *x;
    if (canvas_dspstate) ugen_stop();
    else sys_gui("pdtk_pd_dsp ON\n");
    ugen_start();
    ugen_start_graph();
    for (x = canvas_list; x; x = x->x_next)
    	if (!x->x_parent->x_supercanvas)
    {
    	t_linetraverser t;
    	t_outconnect *oc;
    	t_gobj *y;
    	t_object *ob;
    	for (y = x->x_list; y; y = y->g_next)
    	    if ((ob = pd_checkobject(&y->g_pd)) && zgetfn(&y->g_pd, dspsym))
    	    	ugen_add(ob);
    	linetraverser_start(&t, x);
    	while (oc = linetraverser_next(&t))
    	{
    	    if (obj_issignaloutlet(t.tr_ob, t.tr_outno))
    	    {
    	    	ugen_connect(t.tr_ob, t.tr_outno, t.tr_ob2, t.tr_inno);
    	    }
    	}
    }
    ugen_done_graph();
    canvas_dspstate = 1;
}

void canvas_stop_dsp(void)
{
    if (canvas_dspstate)
    {
    	ugen_stop();
    	sys_gui("pdtk_pd_dsp OFF\n");
    	canvas_dspstate = 0;
    }
}

int canvas_getdspstate(void)
{
    return(canvas_dspstate);
}

void glob_dsp(void *dummy, t_symbol *s, int argc, t_atom *argv)
{
    int newstate;
    if (argc)
    {
    	newstate = atom_getintarg(0, argc, argv);
    	if (newstate && !canvas_dspstate)
    	    canvas_start_dsp();
    	else if (!newstate && canvas_dspstate)
    	    canvas_stop_dsp();
    }
    else post("dsp state %d", canvas_dspstate);
}

/* ------------------------ fields ------------------------ */

int canvas_template_size(t_canvas *x)
{
    int size = 0;
    t_gobj *g;
    if (!x) return (sizeof(t_float));
    for (g = x->x_glist.gl_list; g; g = g->g_next)
    {
    	if (g->g_pd == field_class)
    	    size += field_size((t_field *)g);
    }
    return (size);
}

extern t_field *field_default(void);

t_field *canvas_find_field(t_canvas *x, t_symbol *name, int *onsetp)
{
    int onset = 0;
    t_gobj *g;
    if (!x)
    {
    	if (strcmp(name->s_name, "y")) return (0);
    	*onsetp = 0;
    	return (field_default());
    }
    for (g = x->x_glist.gl_list; g; g = g->g_next)
    {
    	if (g->g_pd == field_class)
    	{
    	    if (field_name((t_field *)g) == name)
    	    {
    	    	*onsetp = onset;
    	    	return ((t_field *)g);
    	    }
    	}
    	onset += field_size((t_field *)g);
    }
    return (0);
}

/* ------------------------------- setup routine ------------------------ */

void g_canvas_setup()
{
    canvas_class = class_new(gensym("canvas"), canvas_new, canvas_free,
    	sizeof(t_canvas), CLASS_PD, A_GIMME, 0);

/* -------------------------- objects ----------------------------- */
    class_addmethod(canvas_class, canvas_obj, gensym("obj"), A_GIMME, A_NULL);
    class_addmethod(canvas_class, canvas_msg, gensym("msg"), A_GIMME, A_NULL);
    class_addmethod(canvas_class, canvas_intatom, gensym("intatom"),
    	A_GIMME, A_NULL);
    class_addmethod(canvas_class, canvas_floatatom, gensym("floatatom"),
    	A_GIMME, A_NULL);
    class_addmethod(canvas_class, canvas_symbolatom, gensym("symbolatom"),
    	A_GIMME, A_NULL);
    class_addmethod(canvas_class, glist_text, gensym("text"), A_GIMME, A_NULL);
    class_addmethod(canvas_class, glist_graph, gensym("graph"),
    	A_GIMME, A_NULL);
    class_addmethod(canvas_class, canvas_array, gensym("array"), A_NULL);

/* ---------------------- methods for reading files ------------------ */
    class_addmethod(canvas_class, canvas_font, gensym("font"), A_SYMBOL, A_NULL);
    class_addmethod(canvas_class, canvas_connect, gensym("connect"),
    	A_INT, A_INT, A_INT, A_INT, A_NULL);
    class_addmethod(canvas_class, canvas_restore, gensym("restore"),
    	A_GIMME, 0);

/* ------------------------ events ---------------------------------- */
    class_addmethod(canvas_class, canvas_click, gensym("click"),
    	A_INT, A_INT, A_INT, A_INT, A_NULL);
    class_addmethod(canvas_class, canvas_mouseup, gensym("mouseup"),
    	A_INT, A_INT, A_INT, A_INT, A_INT, A_NULL);
    class_addmethod(canvas_class, canvas_key, gensym("key"),
    	A_INT, A_NULL);
    class_addmethod(canvas_class, canvas_motion, gensym("motion"),
    	A_INT, A_INT, A_NULL);

/* ------------------------ window stuff --------------------------- */
    class_addmethod(canvas_class, canvas_print, gensym("print"),
    	A_SYMBOL, A_NULL);
    class_addmethod(canvas_class, canvas_menusave, gensym("menusave"), 0);
    class_addmethod(canvas_class, canvas_menusaveas, gensym("menusaveas"), 0);
    class_addmethod(canvas_class, canvas_menuclose, gensym("menuclose"),
    	A_DEFINT, 0);
    class_addmethod(canvas_class, canvas_saveto, gensym("saveto"), 
    	A_SYMBOL, A_SYMBOL, 0);
    class_addmethod(canvas_class, canvas_cut, gensym("cut"), A_NULL);
    class_addmethod(canvas_class, canvas_copy, gensym("copy"), A_NULL);
    class_addmethod(canvas_class, canvas_paste, gensym("paste"), A_NULL);
    class_addmethod(canvas_class, canvas_duplicate, gensym("duplicate"),
    	A_NULL);
    class_addmethod(canvas_class, canvas_texteditor, gensym("texteditor"),
    	A_NULL);
    class_addmethod(canvas_class, canvas_editmode, gensym("editmode"),
    	A_DEFINT, A_NULL);
    class_addmethod(canvas_class, canvas_print, gensym("print"),
    	A_SYMBOL, A_NULL);
    class_addmethod(canvas_class, canvas_pop, gensym("pop"), A_DEFINT, A_NULL);
    class_addmethod(canvas_class, canvas_loadbang, gensym("loadbang"), A_NULL);

/* -------------- subcanvases, currently named "pd" ------------------ */
    subcanvas_class = class_new(gensym("pd"), subcanvas_new, 
    	subcanvas_free, sizeof(t_subcanvas), CLASS_NOINLET, A_DEFSYM, 0);

    class_addcreator(subcanvas_new, gensym("page"),  A_DEFSYM, 0);
    class_addmethod(subcanvas_class, subcanvas_saveto, gensym("saveto"),
    	A_CANT, 0);
    class_addmethod(subcanvas_class, subcanvas_click, gensym("click"),
    	A_INT, A_INT, A_INT, 0);

/* -------------- obsolete ------------------ */
    class_addmethod(canvas_class, canvas_intatom, gensym("atom"),
    	A_GIMME, A_NULL);

/* -------------- copy buffer ------------------ */
    copy_binbuf = binbuf_new();
}
