/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

/* a header for canvas objects that hold lists of t_gobjs, such as
*  canvases themselves or graphs
*/

/* --------------------- geometry ---------------------------- */
#define IOWIDTH 7
#define IOMIDDLE 3
#define IOTOPMARGIN 1
#define IOBOTTOMMARGIN 1

/* ----------------------- data ------------------------------- */

extern struct _rtext;
#define t_rtext struct _rtext

extern struct _field;
#define t_field struct _field

typedef struct _selection
{
    t_gobj *sel_what;
    struct _selection *sel_next;
} t_selection;

typedef struct _editor
{
    t_rtext *e_rtext;	    	    /* text responder linked list */
    t_selection *e_selection;  	    /* head of the selection list */
    t_rtext *e_textedfor;   	    /* the rtext if any that we are editing */
    t_gobj *e_grab;	    	    /* object being "dragged" */
    t_binbuf *e_connectbuf;	    /* connections to deleted objects */
    t_binbuf *e_deleted;    	    /* last stuff we deleted */
    short e_xwas;   	    	    /* xpos on last mousedown or motion event */
    short e_ywas;   	    	    /* ypos, similarly */
    unsigned int e_onmotion: 3;     /* action to take on motion */
    unsigned int e_lastmoved: 1;    /* one if mouse has moved since click */
    unsigned int e_textdirty: 1;    /* one if e_textedfor has changed */
} t_editor;

#define MA_NONE    0 	/* do nothing on mouse motion */
#define MA_MOVE    1	/* drag the selection around */
#define MA_CONNECT 2	/* make a connection */
#define MA_REGION  3	/* selection region */
#define MA_PASSOUT 4	/* send on to e_grab */

typedef struct _glist
{
    t_gobj gl_gobj;
    t_gobj *gl_list;
    struct _glist *gl_owner;
    float gl_px1;   	    	/* bounding rectangle in parent's coords */
    float gl_py1;
    float gl_px2;
    float gl_py2;
    float gl_x1;    	    	/* ... and in our own coordinates */
    float gl_y1;
    float gl_x2;
    float gl_y2;
    t_editor *gl_editor;
    t_symbol *gl_name;
} t_glist;

#define gl_pd gl_gobj.g_pd

/* function types used to define graphical behavior for gpbjs, a bit like X
widgets.  These aren't Pd methods because Pd's typechecking can't specify the
types of pointer arguments.  Also it's more convenient this way, since
every "patchable" object can just get the "text" behaviors. */

    	/* Call this to get a gobj's bounding rectangle: */
typedef void (*t_getrectfn)(t_gobj *x, struct _glist *glist,
    int *x1, int *y1, int *x2, int *y2);
    	/* and this to displace a gobj: */
typedef void (*t_displacefn)(t_gobj *x, struct _glist *glist, int dx, int dy);
    	/* change color to show selection: */
typedef void (*t_selectfn)(t_gobj *x, struct _glist *glist, int state);
    	/* change appearance to show activation/deactivation: */
typedef void (*t_activatefn)(t_gobj *x, struct _glist *glist, int state);
    	/*  deletion */
typedef void (*t_deletefn)(t_gobj *x, struct _glist *glist);
    	/*  making visible or invisible */
typedef void (*t_visfn)(t_gobj *x, struct _glist *glist, int flag);
    	/*  save to a binbuf */
typedef void (*t_savefn)(t_gobj *x, t_binbuf *b);
    	/* ... and later, resizing; getting/setting color... */

struct _widgetbehavior
{
    t_getrectfn w_getrectfn;
    t_displacefn w_displacefn;
    t_selectfn w_selectfn;
    t_activatefn w_activatefn;
    t_deletefn w_deletefn;
    t_visfn w_visfn;
    t_savefn w_savefn;
};

extern t_canvas *canvas_editing;    /* last canvas to start text edting */ 

/* ------------------- functions on any gobj ----------------------------- */
void gobj_getrect(t_gobj *x, t_glist *owner, int *x1, int *y1,
    int *x2, int *y2);
void gobj_displace(t_gobj *x, t_glist *owner, int dx, int dy);
void gobj_select(t_gobj *x, t_glist *owner, int state);
void gobj_activate(t_gobj *x, t_glist *owner, int state);
void gobj_delete(t_gobj *x, t_glist *owner);
void gobj_vis(t_gobj *x, t_glist *glist, int flag);
void gobj_save(t_gobj *x, t_binbuf *b);

/* -------------------- functions on glists --------------------- */
void glist_add(t_glist *x, t_gobj *g);
char *glist_getfont(t_glist *x);
t_canvas *glist_getcanvas(t_glist *x);
void glist_init(t_glist *x);
int glist_isselected(t_glist *x, t_gobj *y);
void glist_select(t_glist *x, t_gobj *y);
void glist_deselect(t_glist *x, t_gobj *y);
void glist_noselect(t_glist *x);
void glist_selectall(t_glist *x);
void glist_delete(t_glist *x, t_gobj *y);
void glist_retext(t_glist *x, t_text *y);
void glist_grab(t_glist *x, t_gobj *y, t_int xpos, t_int ypos);
int  glist_isvisible(t_glist *x);
t_glist *glist_findgraph(t_glist *x, float x1, float x2, float y1, float y2);

float glist_pixelstox(t_glist *x, float xpix);
float glist_pixelstoy(t_glist *x, float ypix);
float glist_xtopixels(t_glist *x, float xval);
float glist_ytopixels(t_glist *x, float yval);

/* -------------------- functions on texts ------------------------- */
void text_setto(t_text *x, t_glist *glist, char *buf, int bufsize);
void text_drawborder(t_text *x, t_glist *glist, t_symbol *tag,
    int width, int height, int firsttime);
void text_eraseborder(t_text *x, t_glist *glist, t_symbol *tag);

/* -------------------- functions on rtexts ------------------------- */
t_rtext *rtext_new(t_glist *glist, t_text *who, t_rtext *next);
t_rtext *rtext_remove(t_rtext *first, t_rtext *x);
t_rtext *glist_findrtext(t_glist *gl, t_text *who);
int rtext_height(t_rtext *x);
void rtext_displace(t_rtext *x, int dx, int dy);
void rtext_select(t_rtext *x, int state);
void rtext_activate(t_rtext *x, int state);
void rtext_cutoff(t_rtext *x);
void rtext_key(t_rtext *x, int key);
void rtext_retext(t_rtext *x);
int rtext_width(t_rtext *x);
int rtext_height(t_rtext *x);
t_symbol *rtext_getsym(t_rtext *x);
void rtext_gettext(t_rtext *x, char **buf, int *bufsize);

/* ---------------------- functions on subcanvases --------------------- */

t_inlet *canvas_addinlet(t_canvas *x, t_pd *who);
void canvas_rminlet(t_canvas *x, t_inlet *ip);
t_outlet *canvas_addoutlet(t_canvas *x, t_pd *who);
void canvas_rmoutlet(t_canvas *x, t_outlet *op);

/* -------------------- functions on canvases ------------------------ */
extern t_class *canvas_class;
void canvas_vistext(t_canvas *x, t_text *y);
void canvas_fixlinesfor(t_canvas *x, t_text *text);
void canvas_deletelinesfor(t_canvas *x, t_text *text);
void canvas_stowconnections(t_canvas *x);
void canvas_restoreconnections(t_canvas *x);
int canvas_template_size(t_canvas *x);
t_field *canvas_find_field(t_canvas *x, t_symbol *name, int *onset);
t_canvas *canvas_getcurrent(void);
t_canvas *canvas_setcurrent(t_canvas *x);
t_canvas *canvas_unsetcurrent(t_canvas *x);
t_canvas *canvas_getrootfor(t_canvas *x);
void canvas_dirty(t_canvas *x, t_int n);

typedef int (*t_canvasapply)(t_canvas *x, t_int x1, t_int x2, t_int x3);

t_int *canvas_recurapply(t_canvas *x, t_canvasapply *fn,
    t_int x1, t_int x2, t_int x3);

/* -------------------- functions on fields --------------------------- */
extern t_class *field_class;
int field_size(t_field *x);
t_symbol *field_type(t_field *x);
t_symbol *field_name(t_field *x);

/* -------------------- functions on arrays --------------------------- */
void array_redraw(t_array *x);
t_canvas *array_template(t_array *x);
int array_npoints(t_array *x);
char *array_vec(t_array *x);
