/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

#include "m_imp.h"
#include <stdio.h>
#include <string.h>

t_int atom_getint(t_atom *a)
{
    if (a->a_type == A_INT) return (a->a_w.w_int);
    else if (a->a_type == A_FLOAT) return (a->a_w.w_float);
    else return (0);
}

t_float atom_getfloat(t_atom *a)
{
    if (a->a_type == A_FLOAT) return (a->a_w.w_float);
    else if (a->a_type == A_INT) return (a->a_w.w_int);
    else return (0);
}

t_symbol *atom_getsym(t_atom *a)  /* LATER think about this more carefully */
{
    char buf[30];
    if (a->a_type == A_SYMBOL) return (a->a_w.w_symbol);
    else if (a->a_type == A_FLOAT) return (&s_float);
    else return (&s_int);
}

t_symbol *atom_gensym(t_atom *a)  /* this works  better for graph labels */
{
    char buf[30];
    if (a->a_type == A_SYMBOL) return (a->a_w.w_symbol);
    else if (a->a_type == A_FLOAT)
    	sprintf(buf, "%g", a->a_w.w_float);
    else sprintf(buf, "%d", a->a_w.w_int);
    return (gensym(buf));
}

t_int atom_getintarg(int which, int argc, t_atom *argv)
{
    if (argc <= which) return (0);
    argv += which;
    if (argv->a_type == A_INT) return (argv->a_w.w_int);
    else if (argv->a_type == A_FLOAT) return (argv->a_w.w_float);
    else return (0);
}

t_float atom_getfloatarg(int which, int argc, t_atom *argv)
{
    if (argc <= which) return (0);
    argv += which;
    if (argv->a_type == A_FLOAT) return (argv->a_w.w_float);
    else if (argv->a_type == A_INT) return (argv->a_w.w_int);
    else return (0);
}

t_symbol *atom_getsymarg(int which, int argc, t_atom *argv)
{
    if (argc <= which) return (&s_);
    argv += which;
    if (argv->a_type == A_SYMBOL) return (argv->a_w.w_symbol);
    else return (&s_);
}

/* convert an atom into a string, in the reverse sense of binbuf_text (q.v.)
* special attention is paid to symbols containing the special characters
* ';', ',', '$', and '\'; these are quoted with a preceding '\', except that
* the '$' only gets quoted at the beginning of the string.
*/

void atom_string(t_atom *a, char *buf, int bufsize)
{
    char tbuf[30];
    switch(a->a_type)
    {
    case A_SEMI: strcpy(buf, ";"); break;
    case A_COMMA: strcpy(buf, ","); break;
    case A_INT:
    	sprintf(tbuf, "%d", a->a_w.w_int);
    	if (strlen(tbuf) < bufsize-1) strcpy(buf, tbuf);
    	else if (a->a_w.w_int < 0) strcpy(buf, "-");
    	else  strcat(buf, "+");
    	break;
    case A_FLOAT:
    	sprintf(tbuf, "%g", a->a_w.w_float);
    	if (!strchr(tbuf, '.')) sprintf(tbuf, "%.1f", a->a_w.w_float);
    	if (strlen(tbuf) < bufsize-1) strcpy(buf, tbuf);
    	else if (a->a_w.w_float < 0) strcpy(buf, "-");
    	else  strcat(buf, "+");
    	break;
    case A_SYMBOL:
    {
    	char *sp;
    	int len, quote;
    	for (sp = a->a_w.w_symbol->s_name, len = 0, quote = 0; *sp; sp++, len++)
    	    if (*sp == ';' || *sp == ',' || *sp == '\\' || *sp == '$')
    	    	quote = 1;
    	if (quote)
    	{
    	    char *bp = buf, *ep = buf + (bufsize-2);
    	    sp = a->a_w.w_symbol->s_name;
    	    while (bp < ep && *sp)
    	    {
    	    	if (*sp == ';' || *sp == ',' || *sp == '\\' ||
    	    	    *sp == '$' && bp == buf)
    	    	    	*bp++ = '\\';
    	    	*bp++ = *sp++;
    	    }
    	    if (*sp) *bp++ = '*';
    	    *bp = 0;
    	    /* post("quote %s -> %s", a->a_w.w_symbol->s_name, buf); */
    	}
    	else
    	{
    	    if (len < bufsize-1) strcpy(buf, a->a_w.w_symbol->s_name);
    	    else
    	    {
    		strncpy(buf, a->a_w.w_symbol->s_name, bufsize - 2);
    		strcpy(a->a_w.w_symbol->s_name + (bufsize - 2), "*");
    	    }
	}
    }
    	break;
    case A_DOLLAR:
    	sprintf(buf, "$%d", a->a_w.w_int);
    	break;
    case A_DOLLSYM:
    	sprintf(buf, "$%s", a->a_w.w_symbol->s_name);
    	break;
    default:
    	bug("atom_string");
    }
}
