/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#ifdef IRIX
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <bstring.h>
#include <sys/time.h>
#include <errno.h>
#endif
#ifdef NT
#include <winsock.h>
#include <io.h>
#endif
#include "tk.h"
#include "t_tk.h"

void tcl_mess(char *s);

/***************** the socket setup code ********************/

static int portno = 5400;

void pdgui_setsock(int port)
{
    portno = port;
}

static void pdgui_sockerror(char *s)
{
#ifdef NT
    int err = WSAGetLastError();
#endif
#ifdef IRIX
    int err = errno;
#endif

    fprintf(stderr, "%s: %s (%d)\n", s, strerror(err), err);
    tcl_mess("exit\n");
    exit(1);
}

static int sockfd;

void pdgui_setupsocket(void)
{
    struct sockaddr_in server;
    struct hostent *hp;

#ifdef NT
    short version = MAKEWORD(2, 0);
    WSADATA nobby;

    if (WSAStartup(version, &nobby)) pdgui_sockerror("setup");
#endif

    /* create a socket */
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) pdgui_sockerror("socket");

    /* connect socket using hostname provided in command line */
    server.sin_family = AF_INET;
    hp = gethostbyname("localhost");
    if (hp == 0)
    {
	fprintf(stderr, "bad host?\n");
	exit(1);
    }
    memcpy((char *)&server.sin_addr, (char *)hp->h_addr, hp->h_length);

    /* assign client port number */
    server.sin_port = portno;

	/* try to connect */
    if (connect(sockfd, (struct sockaddr *) &server, sizeof (server)) < 0)
    	pdgui_sockerror("connecting stream socket");

}

/**************************** commands ************************/
static char *pdgui_path;

/* The "pd" command, which cats its args together and throws the result
* at the Pd interpreter.
*/
#define MAXWRITE 1024

static int pdCmd(ClientData cd, Tcl_Interp *interp, int argc,  char **argv)
{
    if (argc == 2)
    {
    	int n = strlen(argv[1]);
	if (send(sockfd, argv[1], n, 0) < n)
	{
    	    perror("stdout");
    	    tcl_mess("exit\n");
	}
    }
    else
    {
    	int i;
    	char buf[MAXWRITE];
    	buf[0] = 0;
    	for (i = 1; i < argc; i++)
    	{
    	    if (strlen(argv[i]) + strlen(buf) + 2 > MAXWRITE)
	    {
		interp->result = "pd: arg list too long";
		return (TCL_ERROR);	
	    }
	    if (i > 1) strcat(buf, " ");
	    strcat(buf, argv[i]);
	}
	if (send(sockfd, buf, strlen(buf), 0) < 0)
	{
    	    perror("stdout");
    	    tcl_mess("exit\n");
	}
    }
    return (TCL_OK);
}


/* The "pd_guifile" command, which evaluates a file in tk, whose name is
* given relative to the "pd" directory.  I wish I could find a
* better name than "guifile".
*/
void pdgui_doevalfile(Tcl_Interp *interp, char *s);

static int pd_guifileCmd(ClientData cd, Tcl_Interp *interp,
    int argc, char **argv)
{
    if (argc < 2)
    {
	interp->result = "pd: wants at least one argument";
	return (TCL_ERROR);	
    }
    pdgui_doevalfile(interp, argv[1]);
    return (TCL_OK);
}


/* The "pd_getdir" command, which returns the base directory of the PD
* release. 
*/

static int pd_getdirCmd(ClientData cd, Tcl_Interp *interp,
    int argc, char **argv)
{
    interp->result = pdgui_path;
    return (TCL_OK);
}


/* The "pd_suck" command, which polls the socket. 
*/
#define SOCKSIZE 10000

static char pd_tkbuf[SOCKSIZE+1];
int pd_spillbytes = 0;

static int pd_sucksocketCmd(ClientData cd, Tcl_Interp *interp,
    int argc, char **argv)
{
    int ngot;
    fd_set readset, writeset, exceptset;
    struct timeval timout;

    timout.tv_sec = 0;
    timout.tv_usec = 0;
    FD_ZERO(&writeset);
    FD_ZERO(&readset);
    FD_ZERO(&exceptset);
    FD_SET(sockfd, &readset);
    FD_SET(sockfd, &exceptset);
    
    if (select(sockfd+1, &readset, &writeset, &exceptset, &timout) < 0)
    	perror("select");
    if (FD_ISSET(sockfd, &exceptset) || FD_ISSET(sockfd, &readset))
    {
	int ret;
	ret = recv(sockfd, pd_tkbuf + pd_spillbytes,
	    SOCKSIZE - pd_spillbytes, 0);
	if (ret < 0) pdgui_sockerror("socket receive error");
	else if (ret == 0)
	{
    	    fprintf(stderr, "pd: gui died\n");
    	    tcl_mess("exit\n");
	}
	else
	{
	    char *lastcr = 0, *bp = pd_tkbuf + pd_spillbytes + ret;
	    while (bp > pd_tkbuf)
	    {
	    	bp--;
	    	if (*bp == '\n' && (bp == pd_tkbuf || bp[-1] != '\\'))
	    	{
	    	    lastcr = bp;
	    	    break;
	    	}
	    }
	    if (lastcr)
	    {
	    	int xtra = pd_tkbuf + pd_spillbytes + ret - (lastcr+1);
	    	char bashwas = lastcr[1];
	    	lastcr[1] = 0;
	    	tcl_mess(pd_tkbuf);
	    	lastcr[1] = bashwas;
	    	if (xtra)
	    	{
	    	    /* fprintf(stderr, "x %d\n", xtra); */
	    	    memmove(pd_tkbuf, lastcr+1, xtra);
	    	}
	    	pd_spillbytes = xtra;
	    }
	    else
	    {
	    	pd_spillbytes += ret;
	    }
	}
    }
    
    return (TCL_OK);
}

/***********  "c" level access to tk functions. ******************/

static Tcl_Interp *tk_myinterp;

void tcl_mess(char *s)
{
    int result;
    result = Tcl_Eval(tk_myinterp,  s);
    if (result != TCL_OK)
    {
    	if (*tk_myinterp->result) printf("%s\n",  tk_myinterp->result);
    }
}

/* LATER should do a bounds check -- but how do you get printf to do that? */
void tcl_vmess(char *fmt, ...)
{
    int result, i;
    char buf[MAXWRITE];
    va_list ap;
    
    va_start(ap, fmt);

    vsprintf(buf, fmt, ap);
    result = Tcl_Eval(tk_myinterp, buf);
    if (result != TCL_OK)
    {
    	if (*tk_myinterp->result) printf("%s\n",  tk_myinterp->result);
    }
    va_end(ap);
}

void pdgui_doevalfile(Tcl_Interp *interp, char *s)
{
    char buf[GUISTRING];
#ifdef IRIX
    strcpy(buf, pdgui_path);
    strcat(buf, "src/");
    strcat(buf, s);
#endif
#ifdef NT
    strcpy(buf, s);
#endif
    if (Tcl_EvalFile(interp, buf) != TCL_OK)
    	tcl_mess("puts pdgui_evalfile\n");
    	
}

void pdgui_evalfile(char *s)
{
    pdgui_doevalfile(tk_myinterp, s);
}


void pdgui_startup(Tcl_Interp *interp)
{

    	/* save pointer to the main interpreter */
    tk_myinterp = interp;

    	/* add our own TK commands */
    Tcl_CreateCommand(interp, "pd",  pdCmd, (ClientData)NULL, 
	(Tcl_CmdDeleteProc *)NULL); 
    Tcl_CreateCommand(interp, "pd_guifile",  pd_guifileCmd, (ClientData)NULL, 
	(Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateCommand(interp, "pd_getdir",  pd_getdirCmd, (ClientData)NULL, 
	(Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateCommand(interp, "pd_suck",  pd_sucksocketCmd,
    	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

    pdgui_setupsocket();
    
	/* read in the startup file */
#ifdef IRIX
    pdgui_evalfile("u_main.tk");
#endif
#ifdef NT
    pdgui_path = "/inetpub/ftproot/pd";
    tcl_mess("source /inetpub/ftproot/pd/src/u_main.tk\n");
#endif
}

void pdgui_setname(char *s)
{
    char *t;
    char *str;
    int n;
    if (t = strrchr(s, '/')) str = s, n = (t-s) + 1;
    else str = "./", n = 2;
    if (n > GUISTRING-100) n = GUISTRING-100;
    pdgui_path = malloc(n+9);
    strncpy(pdgui_path, str, n);
    strcpy(pdgui_path+n, "../");
}

#ifdef NT
int Pdtcl_Init(Tcl_Interp *interp)
{
    tk_myinterp = interp;
    pdgui_startup(interp); 
    interp->result = "loaded pdtcl_init";
    return (TCL_OK);
}
#endif
