/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

/* arithmetic.  Ugh! */

#include "m_pd.h"
#include <math.h>


#ifdef NT   	    /* NT doesn't appear to have single-precision ANSI math */
#define sinf sin
#define cosf cos
#define atanf atan
#define atan2f atan2
#define sqrtf sqrt
#define logf log
#define expf exp
#define fabsf fabs
#endif

typedef struct _binop_int
{
    t_object x_obj;
    t_int x_n1;
    t_int x_n2;
} t_binop_int;

typedef struct _binop_float
{
    t_object x_obj;
    t_float x_f1;
    t_float x_f2;
} t_binop_float;

/* ------------------ binop1:  +, -, *, / ----------------------------- */

static void *binop1_new(t_class *fixclass, t_class *floatclass,
    int argc, t_atom *argv)
{
    int floater = 0;
    int intval = 0;
    if (argc > 1) post("%s: extra arguments ignored",
    	class_symbol(fixclass)->s_name);
    else if (!argc || argv->a_type == A_SYMBOL) intval = 0;
    else if (argv->a_type == A_FLOAT) floater = 1;
    else intval = argv->a_w.w_int;
    if (floater)
    {
    	t_binop_float *x = (t_binop_float *)pd_new(floatclass);
    	outlet_new(&x->x_obj, &s_float);
    	floatinlet_new(&x->x_obj, &x->x_f2);
    	x->x_f2 = argv->a_w.w_float;
    	return (x);
    }
    else
    {
    	t_binop_int *x = (t_binop_int *)pd_new(fixclass);
    	outlet_new(&x->x_obj, &s_int);
    	intinlet_new(&x->x_obj, &x->x_n2);
    	x->x_n2 = intval;
    	return (x);
    }
}

/* --------------------- addition ------------------------------- */

static t_class *binop1_plus_int_class, *binop1_plus_float_class;

static void *binop1_plus_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop1_new(binop1_plus_int_class, binop1_plus_float_class,
    	argc, argv));
}

static void binop1_plus_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 + x->x_n2);
}

static void binop1_plus_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) + x->x_n2);
}

static void binop1_plus_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) + x->x_n2);
}

static void binop1_plus_float_bang(t_binop_float *x)
{
    outlet_float(x->x_obj.ob_outlet, x->x_f1 + x->x_f2);
}

static void binop1_plus_float_int(t_binop_float *x, t_int n)
{
    outlet_float(x->x_obj.ob_outlet, (x->x_f1 = n) + x->x_f2);
}

static void binop1_plus_float_float(t_binop_float *x, t_float f)
{
    outlet_float(x->x_obj.ob_outlet, (x->x_f1 = f) + x->x_f2);
}

/* --------------------- subtraction ------------------------------- */

static t_class *binop1_minus_int_class, *binop1_minus_float_class;

static void *binop1_minus_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop1_new(binop1_minus_int_class, binop1_minus_float_class,
    	argc, argv));
}

static void binop1_minus_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 - x->x_n2);
}

static void binop1_minus_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) - x->x_n2);
}

static void binop1_minus_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) - x->x_n2);
}

static void binop1_minus_float_bang(t_binop_float *x)
{
    outlet_float(x->x_obj.ob_outlet, x->x_f1 - x->x_f2);
}

static void binop1_minus_float_int(t_binop_float *x, t_int n)
{
    outlet_float(x->x_obj.ob_outlet, (x->x_f1 = n) - x->x_f2);
}

static void binop1_minus_float_float(t_binop_float *x, t_float f)
{
    outlet_float(x->x_obj.ob_outlet, (x->x_f1 = f) - x->x_f2);
}

/* --------------------- multiplication ------------------------------- */

static t_class *binop1_times_int_class, *binop1_times_float_class;

static void *binop1_times_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop1_new(binop1_times_int_class, binop1_times_float_class,
    	argc, argv));
}

static void binop1_times_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 * x->x_n2);
}

static void binop1_times_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) * x->x_n2);
}

static void binop1_times_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) * x->x_n2);
}

static void binop1_times_float_bang(t_binop_float *x)
{
    outlet_float(x->x_obj.ob_outlet, x->x_f1 * x->x_f2);
}

static void binop1_times_float_int(t_binop_float *x, t_int n)
{
    outlet_float(x->x_obj.ob_outlet, (x->x_f1 = n) * x->x_f2);
}

static void binop1_times_float_float(t_binop_float *x, t_float f)
{
    outlet_float(x->x_obj.ob_outlet, (x->x_f1 = f) * x->x_f2);
}

/* --------------------- division ------------------------------- */

static t_class *binop1_div_int_class, *binop1_div_float_class;

static void *binop1_div_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop1_new(binop1_div_int_class, binop1_div_float_class,
    	argc, argv));
}

static void binop1_div_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet,
    	(x->x_n2 ? x->x_n1 / x->x_n2 : 0));
}

static void binop1_div_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet,
    	(x->x_n2 ? (x->x_n1 = n) / x->x_n2 : 0));
}

static void binop1_div_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet,
    	(x->x_n2 ? (x->x_n1 = f) / x->x_n2 : 0));
}

static void binop1_div_float_bang(t_binop_float *x)
{
    outlet_float(x->x_obj.ob_outlet,
    	(x->x_f2 != 0 ? x->x_f1 / x->x_f2 : 0));
}

static void binop1_div_float_int(t_binop_float *x, t_int n)
{   
    x->x_f1 = n;
    outlet_float(x->x_obj.ob_outlet,
    	(x->x_f2 != 0 ? (x->x_f1 = n) / x->x_f2 : 0));
}

static void binop1_div_float_float(t_binop_float *x, t_float f)
{
    x->x_f1 = f;
    outlet_float(x->x_obj.ob_outlet,
    	(x->x_f2 != 0 ? (x->x_f1 = f) / x->x_f2 : 0));
}

/* ------------------ binop2: ==, !=, >, <, >=, <=. -------------------- */

static void *binop2_new(t_class *fixclass, t_class *floatclass,
    int argc, t_atom *argv)
{
    int floater = 0;
    int intval = 0;
    if (argc > 1) post("%s: extra arguments ignored",
    	class_symbol(fixclass)->s_name);
    else if (!argc || argv->a_type == A_SYMBOL) intval = 0;
    else if (argv->a_type == A_FLOAT) floater = 1;
    else intval = argv->a_w.w_int;
    if (floater)
    {
    	t_binop_float *x = (t_binop_float *)pd_new(floatclass);
    	outlet_new(&x->x_obj, &s_int);
    	floatinlet_new(&x->x_obj, &x->x_f2);
    	x->x_f2 = argv->a_w.w_float;
    	return (x);
    }
    else
    {
    	t_binop_int *x = (t_binop_int *)pd_new(fixclass);
    	outlet_new(&x->x_obj, &s_int);
    	intinlet_new(&x->x_obj, &x->x_n2);
    	x->x_n2 = intval;
    	return (x);
    }
}

/* --------------------- == ------------------------------- */

static t_class *binop2_ee_int_class, *binop2_ee_float_class;

static void *binop2_ee_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop2_new(binop2_ee_int_class, binop2_ee_float_class,
    	argc, argv));
}

static void binop2_ee_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 == x->x_n2);
}

static void binop2_ee_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) == x->x_n2);
}

static void binop2_ee_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) == x->x_n2);
}

static void binop2_ee_float_bang(t_binop_float *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_f1 == x->x_f2);
}

static void binop2_ee_float_int(t_binop_float *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = n) == x->x_f2);
}

static void binop2_ee_float_float(t_binop_float *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = f) == x->x_f2);
}

/* --------------------- != ------------------------------- */

static t_class *binop2_ne_int_class, *binop2_ne_float_class;

static void *binop2_ne_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop2_new(binop2_ne_int_class, binop2_ne_float_class,
    	argc, argv));
}

static void binop2_ne_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 != x->x_n2);
}

static void binop2_ne_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) != x->x_n2);
}

static void binop2_ne_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) != x->x_n2);
}

static void binop2_ne_float_bang(t_binop_float *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_f1 != x->x_f2);
}

static void binop2_ne_float_int(t_binop_float *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = n) != x->x_f2);
}

static void binop2_ne_float_float(t_binop_float *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = f) != x->x_f2);
}

/* --------------------- > ------------------------------- */

static t_class *binop2_gt_int_class, *binop2_gt_float_class;

static void *binop2_gt_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop2_new(binop2_gt_int_class, binop2_gt_float_class,
    	argc, argv));
}

static void binop2_gt_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 > x->x_n2);
}

static void binop2_gt_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) > x->x_n2);
}

static void binop2_gt_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) > x->x_n2);
}

static void binop2_gt_float_bang(t_binop_float *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_f1 > x->x_f2);
}

static void binop2_gt_float_int(t_binop_float *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = n) > x->x_f2);
}

static void binop2_gt_float_float(t_binop_float *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = f) > x->x_f2);
}

/* --------------------- < ------------------------------- */

static t_class *binop2_lt_int_class, *binop2_lt_float_class;

static void *binop2_lt_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop2_new(binop2_lt_int_class, binop2_lt_float_class,
    	argc, argv));
}

static void binop2_lt_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 < x->x_n2);
}

static void binop2_lt_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) < x->x_n2);
}

static void binop2_lt_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) < x->x_n2);
}

static void binop2_lt_float_bang(t_binop_float *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_f1 < x->x_f2);
}

static void binop2_lt_float_int(t_binop_float *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = n) < x->x_f2);
}

static void binop2_lt_float_float(t_binop_float *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = f) < x->x_f2);
}

/* --------------------- >= ------------------------------- */

static t_class *binop2_ge_int_class, *binop2_ge_float_class;

static void *binop2_ge_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop2_new(binop2_ge_int_class, binop2_ge_float_class,
    	argc, argv));
}

static void binop2_ge_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 >= x->x_n2);
}

static void binop2_ge_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) >= x->x_n2);
}

static void binop2_ge_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) >= x->x_n2);
}

static void binop2_ge_float_bang(t_binop_float *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_f1 >= x->x_f2);
}

static void binop2_ge_float_int(t_binop_float *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = n) >= x->x_f2);
}

static void binop2_ge_float_float(t_binop_float *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = f) >= x->x_f2);
}

/* --------------------- <= ------------------------------- */

static t_class *binop2_le_int_class, *binop2_le_float_class;

static void *binop2_le_new(t_symbol *s, int argc, t_atom *argv)
{
    return (binop2_new(binop2_le_int_class, binop2_le_float_class,
    	argc, argv));
}

static void binop2_le_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 <= x->x_n2);
}

static void binop2_le_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) <= x->x_n2);
}

static void binop2_le_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) <= x->x_n2);
}

static void binop2_le_float_bang(t_binop_float *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_f1 <= x->x_f2);
}

static void binop2_le_float_int(t_binop_float *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = n) <= x->x_f2);
}

static void binop2_le_float_float(t_binop_float *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_f1 = f) <= x->x_f2);
}

/* ------------- binop3: &, |, &&, ||, % -------------------- */

static void *binop3_new(t_class *fixclass, t_int n)
{
    t_binop_int *x = (t_binop_int *)pd_new(fixclass);
    outlet_new(&x->x_obj, &s_int);
    intinlet_new(&x->x_obj, &x->x_n2);
    x->x_n2 = n;
    return (x);
}

/* --------------------------- & ---------------------------- */

static t_class *binop3_ba_int_class;

static void *binop3_ba_new(t_int n)
{
    return (binop3_new(binop3_ba_int_class, n));
}

static void binop2_ba_int_bang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 & x->x_n2);
}

static void binop2_ba_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) & x->x_n2);
}

static void binop2_ba_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) & x->x_n2);
}

/* --------------------------- && ---------------------------- */

static t_class *binop3_la_int_class;

static void *binop3_la_new(t_int n)
{
    return (binop3_new(binop3_la_int_class, n));
}

static void binop2_la_int_lang(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 && x->x_n2);
}

static void binop2_la_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) && x->x_n2);
}

static void binop2_la_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) && x->x_n2);
}

/* --------------------------- | ---------------------------- */

static t_class *binop3_bo_int_class;

static void *binop3_bo_new(t_int n)
{
    return (binop3_new(binop3_bo_int_class, n));
}

static void binop2_bo_int_bong(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 | x->x_n2);
}

static void binop2_bo_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) | x->x_n2);
}

static void binop2_bo_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) | x->x_n2);
}

/* --------------------------- || ---------------------------- */

static t_class *binop3_lo_int_class;

static void *binop3_lo_new(t_int n)
{
    return (binop3_new(binop3_lo_int_class, n));
}

static void binop2_lo_int_long(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 || x->x_n2);
}

static void binop2_lo_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) || x->x_n2);
}

static void binop2_lo_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) || x->x_n2);
}

/* --------------------------- % ---------------------------- */

static t_class *binop3_pc_int_class;

static void *binop3_pc_new(t_int n)
{
    return (binop3_new(binop3_pc_int_class, n));
}

static void binop2_pc_int_pcng(t_binop_int *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n1 % (x->x_n2 ? x->x_n2 : 1));
}

static void binop2_pc_int_int(t_binop_int *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = n) % (x->x_n2 ? x->x_n2 : 1));
}

static void binop2_pc_int_float(t_binop_int *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, (x->x_n1 = f) % (x->x_n2 ? x->x_n2 : 1));
}

/* -------------------- mathematical functions ------------------ */

static t_class *sin_class;   	/* ----------- sin --------------- */

static void *sin_new(void)
{
    t_object *x = (t_object *)pd_new(sin_class);
    outlet_new(x, &s_float);
    return (x);
}

static void sin_float(t_object *x, t_float f)
{
    outlet_float(x->ob_outlet, sinf(f));
}

static t_class *cos_class;   	/* ----------- cos --------------- */

static void *cos_new(void)
{
    t_object *x = (t_object *)pd_new(cos_class);
    outlet_new(x, &s_float);
    return (x);
}

static void cos_float(t_object *x, t_float f)
{
    outlet_float(x->ob_outlet, cosf(f));
}

static t_class *tan_class;   	/* ----------- tan --------------- */

static void *tan_new(void)
{
    t_object *x = (t_object *)pd_new(tan_class);
    outlet_new(x, &s_float);
    return (x);
}

static void tan_float(t_object *x, t_float f)
{
    float c = cosf(f);
    float t = (c == 0 ? 0 : sinf(f)/c);
    outlet_float(x->ob_outlet, t);
}

static t_class *atan_class;   	/* ----------- atan --------------- */

static void *atan_new(void)
{
    t_object *x = (t_object *)pd_new(atan_class);
    outlet_new(x, &s_float);
    return (x);
}

static void atan_float(t_object *x, t_float f)
{
    outlet_float(x->ob_outlet, atanf(f));
}

static t_class *atan2_class;   	/* ----------- atan2 --------------- */

typedef struct _atan2
{
    t_object x_ob;
    float x_y;
} t_atan2;

static void *atan2_new(void)
{
    t_atan2 *x = (t_atan2 *)pd_new(atan2_class);
    floatinlet_new(&x->x_ob, &x->x_y);
    outlet_new(&x->x_ob, &s_float);
    return (x);
}

static void atan2_float(t_atan2 *x, t_float f)
{
    float r = (f == 0 && x->x_y == 0 ? 0 : atan2f(x->x_y, f));
    outlet_float(x->x_ob.ob_outlet, r);
}

static t_class *sqrt_class;   	/* ----------- sqrt --------------- */

static void *sqrt_new(void)
{
    t_object *x = (t_object *)pd_new(sqrt_class);
    outlet_new(x, &s_float);
    return (x);
}

static void sqrt_float(t_object *x, t_float f)
{
    float r = (f > 0 ? sqrtf(f) : 0);
    outlet_float(x->ob_outlet, r);
}

static t_class *log_class;   	/* ----------- log --------------- */

static void *log_new(void)
{
    t_object *x = (t_object *)pd_new(log_class);
    outlet_new(x, &s_float);
    return (x);
}

static void log_float(t_object *x, t_float f)
{
    float r = (f > 0 ? logf(f) : -1000);
    outlet_float(x->ob_outlet, r);
}


static t_class *exp_class;   	/* ----------- exp --------------- */

static void *exp_new(void)
{
    t_object *x = (t_object *)pd_new(exp_class);
    outlet_new(x, &s_float);
    return (x);
}

#define MAXLOG 87.3365
static void exp_float(t_object *x, t_float f)
{
    if (f > MAXLOG) f = MAXLOG;
    outlet_float(x->ob_outlet, expf(f));
}

static t_class *abs_class;   	/* ----------- abs --------------- */

static void *abs_new(void)
{
    t_object *x = (t_object *)pd_new(abs_class);
    outlet_new(x, &s_float);
    return (x);
}

static void abs_float(t_object *x, t_float f)
{
    outlet_float(x->ob_outlet, fabsf(f));
}

void x_arithmetic_setup(void)
{
    binop1_plus_int_class = class_new(gensym("+"), binop1_plus_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop1_plus_int_class, binop1_plus_int_bang);
    class_addint(binop1_plus_int_class, binop1_plus_int_int);
    class_addfloat(binop1_plus_int_class, (t_method)binop1_plus_int_float);

    binop1_plus_float_class = class_new(gensym("+"), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop1_plus_float_class, binop1_plus_float_bang);
    class_addint(binop1_plus_float_class, binop1_plus_float_int);
    class_addfloat(binop1_plus_float_class, (t_method)binop1_plus_float_float);

    binop1_minus_int_class = class_new(gensym("-"), binop1_minus_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop1_minus_int_class, binop1_minus_int_bang);
    class_addint(binop1_minus_int_class, binop1_minus_int_int);
    class_addfloat(binop1_minus_int_class, (t_method)binop1_minus_int_float);

    binop1_minus_float_class = class_new(gensym("-"), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop1_minus_float_class, binop1_minus_float_bang);
    class_addint(binop1_minus_float_class, binop1_minus_float_int);
    class_addfloat(binop1_minus_float_class, (t_method)binop1_minus_float_float);

    binop1_times_int_class = class_new(gensym("*"), binop1_times_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop1_times_int_class, binop1_times_int_bang);
    class_addint(binop1_times_int_class, binop1_times_int_int);
    class_addfloat(binop1_times_int_class, (t_method)binop1_times_int_float);

    binop1_times_float_class = class_new(gensym("*"), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop1_times_float_class, binop1_times_float_bang);
    class_addint(binop1_times_float_class, binop1_times_float_int);
    class_addfloat(binop1_times_float_class, (t_method)binop1_times_float_float);

    binop1_div_int_class = class_new(gensym("/"), binop1_div_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop1_div_int_class, binop1_div_int_bang);
    class_addint(binop1_div_int_class, binop1_div_int_int);
    class_addfloat(binop1_div_int_class, (t_method)binop1_div_int_float);

    binop1_div_float_class = class_new(gensym("/"), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop1_div_float_class, binop1_div_float_bang);
    class_addint(binop1_div_float_class, binop1_div_float_int);
    class_addfloat(binop1_div_float_class, (t_method)binop1_div_float_float);

    	/* ------------------ binop2 ----------------------- */

    binop2_ee_int_class = class_new(gensym("=="), binop2_ee_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop2_ee_int_class, binop2_ee_int_bang);
    class_addint(binop2_ee_int_class, binop2_ee_int_int);
    class_addfloat(binop2_ee_int_class, (t_method)binop2_ee_int_float);

    binop2_ee_float_class = class_new(gensym("=="), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop2_ee_float_class, binop2_ee_float_bang);
    class_addint(binop2_ee_float_class, binop2_ee_float_int);
    class_addfloat(binop2_ee_float_class, (t_method)binop2_ee_float_float);

    binop2_ne_int_class = class_new(gensym("!="), binop2_ne_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop2_ne_int_class, binop2_ne_int_bang);
    class_addint(binop2_ne_int_class, binop2_ne_int_int);
    class_addfloat(binop2_ne_int_class, (t_method)binop2_ne_int_float);

    binop2_ne_float_class = class_new(gensym("!="), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop2_ne_float_class, binop2_ne_float_bang);
    class_addint(binop2_ne_float_class, binop2_ne_float_int);
    class_addfloat(binop2_ne_float_class, (t_method)binop2_ne_float_float);

    binop2_gt_int_class = class_new(gensym(">"), binop2_gt_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop2_gt_int_class, binop2_gt_int_bang);
    class_addint(binop2_gt_int_class, binop2_gt_int_int);
    class_addfloat(binop2_gt_int_class, (t_method)binop2_gt_int_float);

    binop2_gt_float_class = class_new(gensym(">"), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop2_gt_float_class, binop2_gt_float_bang);
    class_addint(binop2_gt_float_class, binop2_gt_float_int);
    class_addfloat(binop2_gt_float_class, (t_method)binop2_gt_float_float);

    binop2_lt_int_class = class_new(gensym("<"), binop2_lt_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop2_lt_int_class, binop2_lt_int_bang);
    class_addint(binop2_lt_int_class, binop2_lt_int_int);
    class_addfloat(binop2_lt_int_class, (t_method)binop2_lt_int_float);

    binop2_lt_float_class = class_new(gensym("<"), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop2_lt_float_class, binop2_lt_float_bang);
    class_addint(binop2_lt_float_class, binop2_lt_float_int);
    class_addfloat(binop2_lt_float_class, (t_method)binop2_lt_float_float);

    binop2_ge_int_class = class_new(gensym(">="), binop2_ge_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop2_ge_int_class, binop2_ge_int_bang);
    class_addint(binop2_ge_int_class, binop2_ge_int_int);
    class_addfloat(binop2_ge_int_class, (t_method)binop2_ge_int_float);

    binop2_ge_float_class = class_new(gensym(">="), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop2_ge_float_class, binop2_ge_float_bang);
    class_addint(binop2_ge_float_class, binop2_ge_float_int);
    class_addfloat(binop2_ge_float_class, (t_method)binop2_ge_float_float);

    binop2_le_int_class = class_new(gensym("<="), binop2_le_new, 0,
    	sizeof(t_binop_int), 0, A_GIMME, 0);
    class_addbang(binop2_le_int_class, binop2_le_int_bang);
    class_addint(binop2_le_int_class, binop2_le_int_int);
    class_addfloat(binop2_le_int_class, (t_method)binop2_le_int_float);

    binop2_le_float_class = class_new(gensym("<="), 0, 0,
    	sizeof(t_binop_float), 0, 0);
    class_addbang(binop2_le_float_class, binop2_le_float_bang);
    class_addint(binop2_le_float_class, binop2_le_float_int);
    class_addfloat(binop2_le_float_class, (t_method)binop2_le_float_float);

    	/* ------------------ binop3 ----------------------- */

    binop3_ba_int_class = class_new(gensym("&"), binop3_ba_new, 0,
    	sizeof(t_binop_int), 0, A_DEFINT, 0);
    class_addbang(binop3_ba_int_class, binop2_ba_int_bang);
    class_addint(binop3_ba_int_class, binop2_ba_int_int);
    class_addfloat(binop3_ba_int_class, (t_method)binop2_ba_int_float);

    binop3_la_int_class = class_new(gensym("&&"), binop3_la_new, 0,
    	sizeof(t_binop_int), 0, A_DEFINT, 0);
    class_addbang(binop3_la_int_class, binop2_la_int_lang);
    class_addint(binop3_la_int_class, binop2_la_int_int);
    class_addfloat(binop3_la_int_class, (t_method)binop2_la_int_float);

    binop3_bo_int_class = class_new(gensym("|"), binop3_bo_new, 0,
    	sizeof(t_binop_int), 0, A_DEFINT, 0);
    class_addbang(binop3_bo_int_class, binop2_bo_int_bong);
    class_addint(binop3_bo_int_class, binop2_bo_int_int);
    class_addfloat(binop3_bo_int_class, (t_method)binop2_bo_int_float);

    binop3_lo_int_class = class_new(gensym("||"), binop3_lo_new, 0,
    	sizeof(t_binop_int), 0, A_DEFINT, 0);
    class_addbang(binop3_lo_int_class, binop2_lo_int_long);
    class_addint(binop3_lo_int_class, binop2_lo_int_int);
    class_addfloat(binop3_lo_int_class, (t_method)binop2_lo_int_float);

    binop3_pc_int_class = class_new(gensym("%"), binop3_pc_new, 0,
    	sizeof(t_binop_int), 0, A_DEFINT, 0);
    class_addbang(binop3_pc_int_class, binop2_pc_int_pcng);
    class_addint(binop3_pc_int_class, binop2_pc_int_int);
    class_addfloat(binop3_pc_int_class, (t_method)binop2_pc_int_float);

    	/* ------------------- math functions --------------- */
    
    sin_class = class_new(gensym("sin"), sin_new, 0,
    	sizeof(t_object), 0, 0);
    class_addfloat(sin_class, (t_method)sin_float);
    
    cos_class = class_new(gensym("cos"), cos_new, 0,
    	sizeof(t_object), 0, 0);
    class_addfloat(cos_class, (t_method)cos_float);
    
    tan_class = class_new(gensym("tan"), tan_new, 0,
    	sizeof(t_object), 0, 0);
    class_addfloat(tan_class, (t_method)tan_float);

    atan_class = class_new(gensym("atan"), atan_new, 0,
    	sizeof(t_object), 0, 0);
    class_addfloat(atan_class, (t_method)atan_float);

    atan2_class = class_new(gensym("atan2"), atan2_new, 0,
    	sizeof(t_atan2), 0, 0);
    class_addfloat(atan2_class, (t_method)atan2_float);    

    sqrt_class = class_new(gensym("sqrt"), sqrt_new, 0,
    	sizeof(t_object), 0, 0);
    class_addfloat(sqrt_class, (t_method)sqrt_float);

    log_class = class_new(gensym("log"), log_new, 0,
    	sizeof(t_object), 0, 0);
    class_addfloat(log_class, (t_method)log_float);    

    exp_class = class_new(gensym("exp"), exp_new, 0,
    	sizeof(t_object), 0, 0);
    class_addfloat(exp_class, (t_method)exp_float);

    abs_class = class_new(gensym("abs"), abs_new, 0,
    	sizeof(t_object), 0, 0);
    class_addfloat(abs_class, (t_method)abs_float);    

}


