/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

/* connective objects */

#include "m_pd.h"
#include <string.h>
extern t_pd *newest;

/* -------------------------- int ------------------------------ */
static t_class *pdint_class;

typedef struct _pdint
{
    t_object x_obj;
    t_int x_n;
} t_pdint;

static void pdint_bang(t_pdint *x)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n);
}

static void pdint_int(t_pdint *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n = n);
}

static void pdint_float(t_pdint *x, t_float f)
{
    outlet_int(x->x_obj.ob_outlet, x->x_n = f);
}

static void *pdint_new(t_pd *dummy, t_int n)
{
    t_pdint *x = (t_pdint *)pd_new(pdint_class);
    x->x_n = n;
    outlet_new(&x->x_obj, &s_int);
    intinlet_new(&x->x_obj, &x->x_n);
    newest = &x->x_obj.ob_pd;
}

void pdint_setup(void)
{
    pdint_class = class_new(gensym("int"), pdint_new, 0,
    	sizeof(t_pdint), 0, A_DEFINT, 0);
    class_addbang(pdint_class, pdint_bang);
    class_addint(pdint_class, pdint_int);
    class_addfloat(pdint_class, (t_method)pdint_float);
}

/* -------------------------- float ------------------------------ */
static t_class *pdfloat_class;

typedef struct _pdfloat
{
    t_object x_obj;
    t_float x_f;
} t_pdfloat;

static void *pdfloat_new(t_pd *dummy, t_float f)
{
    t_pdfloat *x = (t_pdfloat *)pd_new(pdfloat_class);
    x->x_f = f;
    outlet_new(&x->x_obj, &s_float);
    floatinlet_new(&x->x_obj, &x->x_f);
    newest = &x->x_obj.ob_pd;
}

static void pdfloat_bang(t_pdfloat *x)
{
    outlet_float(x->x_obj.ob_outlet, x->x_f);
}

static void pdfloat_int(t_pdfloat *x, t_int n)
{
    outlet_float(x->x_obj.ob_outlet, x->x_f = n);
}

static void pdfloat_float(t_pdfloat *x, t_float f)
{
    outlet_float(x->x_obj.ob_outlet, x->x_f = f);
}

void pdfloat_setup(void)
{
    pdfloat_class = class_new(gensym("float"), (t_newmethod)pdfloat_new, 0,
    	sizeof(t_pdfloat), 0, A_DEFFLOAT, 0);
    class_addbang(pdfloat_class, pdfloat_bang);
    class_addint(pdfloat_class, pdfloat_int);
    class_addfloat(pdfloat_class, (t_method)pdfloat_float);
}

/* -------------------- send ------------------------------ */

static t_class *send_class;

typedef struct _send
{
    t_object x_obj;
    t_symbol *x_sym;
} t_send;

static void send_bang(t_send *x)
{
    if (x->x_sym->s_thing) pd_bang(x->x_sym->s_thing);
}

static void send_int(t_send *x, t_int n)
{
    if (x->x_sym->s_thing) pd_int(x->x_sym->s_thing, n);
}

static void send_float(t_send *x, t_float f)
{
    if (x->x_sym->s_thing) pd_float(x->x_sym->s_thing, f);
}

static void send_symbol(t_send *x, t_symbol *s)
{
    if (x->x_sym->s_thing) pd_symbol(x->x_sym->s_thing, s);
}

static void send_list(t_send *x, t_symbol *s, int argc, t_atom *argv)
{
    if (x->x_sym->s_thing) pd_list(x->x_sym->s_thing, s, argc, argv);
}

static void send_anything(t_send *x, t_symbol *s, int argc, t_atom *argv)
{
    if (x->x_sym->s_thing) typedmess(x->x_sym->s_thing, s, argc, argv);
}

static void *send_new(t_symbol *s)
{
    t_send *x = (t_send *)pd_new(send_class);
    x->x_sym = s;
    return (x);
}

static void send_setup(void)
{
    send_class = class_new(gensym("send"), send_new, 0,
    	sizeof(t_send), 0, A_DEFSYM, 0);
    class_addcreator(send_new, gensym("s"), A_DEFSYM, 0);
    class_addbang(send_class, send_bang);
    class_addint(send_class, send_int);
    class_addfloat(send_class, (t_method)send_float);
    class_addsymbol(send_class, send_symbol);
    class_addlist(send_class, send_list);
    class_addanything(send_class, send_anything);
}
/* -------------------- receive ------------------------------ */

static t_class *receive_class;

typedef struct _receive
{
    t_object x_obj;
    t_symbol *x_sym;
} t_receive;

static void receive_bang(t_receive *x)
{
    outlet_bang(x->x_obj.ob_outlet);
}

static void receive_int(t_receive *x, t_int n)
{
    outlet_int(x->x_obj.ob_outlet, n);
}

static void receive_float(t_receive *x, t_float f)
{
    outlet_float(x->x_obj.ob_outlet, f);
}

static void receive_symbol(t_receive *x, t_symbol *s)
{
    outlet_symbol(x->x_obj.ob_outlet, s);
}

static void receive_list(t_receive *x, t_symbol *s, int argc, t_atom *argv)
{
    outlet_list(x->x_obj.ob_outlet, s, argc, argv);
}

static void receive_anything(t_receive *x, t_symbol *s, int argc, t_atom *argv)
{
    outlet_anything(x->x_obj.ob_outlet, s, argc, argv);
}

static void *receive_new(t_symbol *s)
{
    t_receive *x = (t_receive *)pd_new(receive_class);
    x->x_sym = s;
    pd_bind(&x->x_obj.ob_pd, s);
    outlet_new(&x->x_obj, 0);
    return (x);
}

static void receive_free(t_receive *x)
{
    pd_unbind(&x->x_obj.ob_pd, x->x_sym);
}

static void receive_setup(void)
{
    receive_class = class_new(gensym("receive"), receive_new, receive_free,
    	sizeof(t_receive), CLASS_NOINLET, A_SYMBOL, 0);
    class_addcreator(receive_new, gensym("r"), A_DEFSYM, 0);
    class_addbang(receive_class, receive_bang);
    class_addint(receive_class, receive_int);
    class_addfloat(receive_class, (t_method)receive_float);
    class_addsymbol(receive_class, receive_symbol);
    class_addlist(receive_class, receive_list);
    class_addanything(receive_class, receive_anything);
}

/* -------------------------- select ------------------------------ */

static t_class *sel1_class;

typedef struct _sel1
{
    t_object x_obj;
    t_int x_n;
    t_outlet *x_outlet1;
    t_outlet *x_outlet2;
} t_sel1;

static void sel1_int(t_sel1 *x, t_int n)
{
    if (n == x->x_n) outlet_bang(x->x_outlet1);
    else outlet_int(x->x_outlet2, n);
}

static void sel1_float(t_sel1 *x, t_float f)
{
    sel1_int(x, (t_int)f);
}

static void sel1_in1(t_sel1 *x, t_intarg n)
{
    x->x_n = n;
}

static t_class *sel2_class;

typedef struct _selectelement
{
    t_int e_n;
    t_outlet *e_outlet;
} t_selectelement;

typedef struct _sel2
{
    t_object x_obj;
    t_int x_nelement;
    t_selectelement *x_vec;
    t_outlet *x_rejectout;
} t_sel2;

static void sel2_int(t_sel2 *x, t_int n)
{
    t_selectelement *e;
    int nelement;
    for (nelement = x->x_nelement, e = x->x_vec; nelement--; e++)
    	if (e->e_n == n)
    {
    	outlet_bang(e->e_outlet);
    	return;
    }
    outlet_int(x->x_rejectout, n);
}

static void sel2_float(t_sel2 *x, t_float f)
{
    sel2_int(x, (t_int)f);
}

static void sel2_free(t_sel2 *x)
{
    freebytes(x->x_vec, x->x_nelement * sizeof(*x->x_vec));
}

static void *select_new(t_symbol *s, int argc, t_atom *argv)
{
    if (argc < 2)
    {
    	t_sel1 *x = (t_sel1 *)pd_new(sel1_class);
    	x->x_n = atom_getintarg(0, argc, argv);
    	x->x_outlet2 = outlet_new(&x->x_obj, &s_int);
    	x->x_outlet1 = outlet_new(&x->x_obj, &s_bang);
    	inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_int, gensym("in1"));
    	return (x);
    }
    else
    {
    	int n;
    	t_selectelement *e;
    	t_sel2 *x = (t_sel2 *)pd_new(sel2_class);
    	x->x_nelement = argc;
    	x->x_vec = (t_selectelement *)getbytes(argc * sizeof(*x->x_vec));
    	x->x_rejectout = outlet_new(&x->x_obj, &s_int);
    	for (n = argc, e = x->x_vec + argc; e--, n--;)
    	{
    	    e->e_outlet = outlet_new(&x->x_obj, &s_bang);
    	    e->e_n = atom_getintarg(n, argc, argv);
    	}
    	return (x);
    }

}

void select_setup(void)
{
    sel1_class = class_new(gensym("select"), 0, 0,
    	sizeof(t_sel1), 0, 0);
    class_addint(sel1_class, sel1_int);
    class_addfloat(sel1_class, (t_method)sel1_float);
    class_addmethod(sel1_class, sel1_in1, gensym("in1"), A_INT, 0);

    sel2_class = class_new(gensym("select"), 0, sel2_free,
    	sizeof(t_sel2), 0, 0);
    class_addint(sel2_class, sel2_int);
    class_addfloat(sel2_class, (t_method)sel2_float);

    class_addcreator(select_new, gensym("select"),  A_GIMME, 0);
    class_addcreator(select_new, gensym("sel"),  A_GIMME, 0);
}

/* -------------------------- route ------------------------------ */

static t_class *route_class;

typedef struct _routeelement
{
    t_int e_n;
    t_outlet *e_outlet;
} t_routeelement;

typedef struct _route
{
    t_object x_obj;
    t_int x_nelement;
    t_routeelement *x_vec;
    t_outlet *x_rejectout;
} t_route;

static void route_list(t_route *x, t_symbol *sel, int argc, t_atom *argv)
{
    t_routeelement *e;
    int nelement, n;
    if (!argc) return;
    else n = atom_getint(argv);
    for (nelement = x->x_nelement, e = x->x_vec; nelement--; e++)
    	if (e->e_n == n)
    {
    	outlet_list(e->e_outlet, 0, argc-1, argv+1);
    	return;
    }
    outlet_list(x->x_rejectout, 0, argc, argv);
}

static void route_free(t_route *x)
{
    freebytes(x->x_vec, x->x_nelement * sizeof(*x->x_vec));
}

static void *route_new(t_symbol *s, int argc, t_atom *argv)
{
    int n;
    t_routeelement *e;
    t_route *x = (t_route *)pd_new(route_class);
    x->x_nelement = argc;
    x->x_vec = (t_routeelement *)getbytes(argc * sizeof(*x->x_vec));
    x->x_rejectout = outlet_new(&x->x_obj, &s_list);
    for (n = argc, e = x->x_vec + argc; e--, n--;)
    {
    	e->e_outlet = outlet_new(&x->x_obj, &s_list);
    	e->e_n = atom_getintarg(n, argc, argv);
    }
    return (x);

}

void route_setup(void)
{
    route_class = class_new(gensym("route"), route_new, route_free,
    	sizeof(t_route), 0, A_GIMME, 0);
    class_addlist(route_class, route_list);
}

/* -------------------------- pack ------------------------------ */

static t_class *pack_class;

typedef struct _pack
{
    t_object x_obj;
    t_int x_n;
    t_atom *x_vec;
} t_pack;

static void *pack_new(t_symbol *s, int argc, t_atom *argv)
{
    t_pack *x = (t_pack *)pd_new(pack_class);
    t_atom defarg[2], *ap;
    int i;
    if (!argc)
    {
    	argv = defarg;
    	argc = 2;
    	SETINT(&defarg[0], 0);
    	SETINT(&defarg[1], 0);
    }
    x->x_n = argc;
    x->x_vec = (t_atom *)getbytes(argc * sizeof(*x->x_vec));
    memcpy(x->x_vec, argv, argc * sizeof(t_atom));
    outlet_new(&x->x_obj, &s_list);
    for (i = argc-1, ap = x->x_vec + i; i--; ap--)
    {
    	if (ap->a_type == A_INT) intinlet_new(&x->x_obj, &ap->a_w.w_int);
    	else if (ap->a_type == A_FLOAT)
    	    floatinlet_new(&x->x_obj, &ap->a_w.w_float);
    	else if (ap->a_type == A_SYMBOL)
    	    symbolinlet_new(&x->x_obj, &ap->a_w.w_symbol);
    }
    return (x);
}

static void pack_bang(t_pack *x)
{
    outlet_list(x->x_obj.ob_outlet, &s_list, x->x_n, x->x_vec);
}

static void pack_int(t_pack *x, t_int n)
{
    if (x->x_vec->a_type == A_INT) x->x_vec->a_w.w_int = n;
    else if (x->x_vec->a_type == A_FLOAT) x->x_vec->a_w.w_float = n;
    else error("pack_int: wanted a symbol");
    outlet_list(x->x_obj.ob_outlet, &s_list, x->x_n, x->x_vec);
}

static void pack_float(t_pack *x, t_float f)
{
    if (x->x_vec->a_type == A_FLOAT) x->x_vec->a_w.w_float = f;
    else if (x->x_vec->a_type == A_INT) x->x_vec->a_w.w_int = f;
    else error("pack_float: wanted a symbol");
    outlet_list(x->x_obj.ob_outlet, &s_list, x->x_n, x->x_vec);
}

static void pack_free(t_pack *x)
{
    freebytes(x->x_vec, x->x_n * sizeof(*x->x_vec));
}

static void pack_setup(void)
{
    pack_class = class_new(gensym("pack"), pack_new, pack_free,
    	sizeof(t_pack), 0, A_GIMME, 0);
    class_addbang(pack_class, pack_bang);
    class_addint(pack_class, pack_int);
    class_addfloat(pack_class, (t_method)pack_float);
}

/* -------------------------- unpack ------------------------------ */

static t_class *unpack_class;

typedef struct unpackout
{
    t_atomtype u_type;
    t_outlet *u_outlet;
} t_unpackout;

typedef struct _unpack
{
    t_object x_obj;
    t_int x_n;
    t_unpackout *x_vec;
} t_unpack;

static void *unpack_new(t_symbol *s, int argc, t_atom *argv)
{
    t_unpack *x = (t_unpack *)pd_new(unpack_class);
    t_atom defarg[2], *ap;
    t_unpackout *u;
    int i;
    if (!argc)
    {
    	argv = defarg;
    	argc = 2;
    	SETINT(&defarg[0], 0);
    	SETINT(&defarg[1], 0);
    }
    x->x_n = argc;
    x->x_vec = (t_unpackout *)getbytes(argc * sizeof(*x->x_vec));
    for (i = argc, ap = argv + i, u = x->x_vec + i; u--, ap--, i--;)
    {
    	u->u_type = ap->a_type;
    	u->u_outlet = outlet_new(&x->x_obj,
    	    (u->u_type == A_INT ? &s_int :
    	    	(u->u_type == A_FLOAT ? &s_float :&s_symbol)));
    }
    return (x);
}

static void unpack_list(t_unpack *x, t_symbol *s, int argc, t_atom *argv)
{
    t_atom *ap;
    t_unpackout *u;
    int i;
    if (argc > x->x_n) argc = x->x_n;
    for (i = argc, u = x->x_vec + i, ap = argv + i; u--, ap--, i--;)
    {
    	if (u->u_type == A_INT) outlet_int(u->u_outlet, atom_getint(ap));
    	else if (u->u_type == A_FLOAT)
    	    outlet_float(u->u_outlet, atom_getfloat(ap));
    	else outlet_symbol(u->u_outlet, atom_getsym(ap));
    }
}

static void unpack_free(t_unpack *x)
{
    freebytes(x->x_vec, x->x_n * sizeof(*x->x_vec));
}

static void unpack_setup(void)
{
    unpack_class = class_new(gensym("unpack"), unpack_new, unpack_free,
    	sizeof(t_unpack), 0, A_GIMME, 0);
    class_addlist(unpack_class, unpack_list);
}

/* -------------------------- trigger ------------------------------ */

static t_class *trigger_class;

typedef struct triggerout
{
    t_atomtype u_type;	/* outlet type, with A_NULL standing for bang and
    	    	    	A_GIMME for list.  */
    t_outlet *u_outlet;
} t_triggerout;

typedef struct _trigger
{
    t_object x_obj;
    t_int x_n;
    t_triggerout *x_vec;
} t_trigger;

static void *trigger_new(t_symbol *s, int argc, t_atom *argv)
{
    t_trigger *x = (t_trigger *)pd_new(trigger_class);
    t_atom defarg[2], *ap;
    t_triggerout *u;
    int i;
    if (!argc)
    {
    	argv = defarg;
    	argc = 2;
    	SETSYM(&defarg[0], &s_bang);
    	SETSYM(&defarg[1], &s_bang);
    }
    x->x_n = argc;
    x->x_vec = (t_triggerout *)getbytes(argc * sizeof(*x->x_vec));
    for (i = argc, ap = argv + i, u = x->x_vec + i; u--, ap--, i--;)
    {
    	t_atomtype thistype = ap->a_type;
    	char c;
    	if (thistype == A_SYMBOL) c = ap->a_w.w_symbol->s_name[0];
    	else if (thistype == A_INT) c = 'i';
    	else if (thistype == A_FLOAT) c = 'f';
    	if (c == 'i')
    	    u->u_type = A_INT, u->u_outlet = outlet_new(&x->x_obj, &s_int);
    	else if (c == 'f')
    	    u->u_type = A_FLOAT, u->u_outlet = outlet_new(&x->x_obj, &s_float);
    	else if (c == 'b')
    	    u->u_type = A_NULL, u->u_outlet = outlet_new(&x->x_obj, &s_bang);
    	else if (c == 'l')
    	    u->u_type = A_GIMME, u->u_outlet = outlet_new(&x->x_obj, &s_list);
    	else
    	    u->u_type = A_SYMBOL, u->u_outlet = outlet_new(&x->x_obj, &s_symbol);
    }
    return (x);
}

static void trigger_list(t_trigger *x, t_symbol *s, int argc, t_atom *argv)
{
    t_triggerout *u;
    int i;
    t_atom at;
    if (!argc)
    {
    	argc = 1;
    	SETINT(&at, 0);
    	argv = &at;
    }
    for (i = x->x_n, u = x->x_vec + i; u--, i--;)
    {
    	if (u->u_type == A_INT) outlet_int(u->u_outlet, atom_getint(argv));
    	else if (u->u_type == A_FLOAT)
    	    outlet_float(u->u_outlet, atom_getfloat(argv));
    	else if (u->u_type == A_NULL) outlet_bang(u->u_outlet);
    	else if (u->u_type == A_GIMME)
    	    outlet_list(u->u_outlet, &s_list, argc, argv);
    	else outlet_symbol(u->u_outlet, atom_getsym(argv));
    }
}

static void trigger_bang(t_trigger *x)
{
    trigger_list(x, 0, 0, 0);
}

static void trigger_int(t_trigger *x, t_int n)
{
    t_atom at;
    SETINT(&at, n);
    trigger_list(x, 0, 1, &at);
}

static void trigger_float(t_trigger *x, t_float f)
{
    t_atom at;
    SETFLOAT(&at, f);
    trigger_list(x, 0, 1, &at);
}

static void trigger_symbol(t_trigger *x, t_symbol *s)
{
    t_atom at;
    SETSYM(&at, s);
    trigger_list(x, 0, 1, &at);
}

static void trigger_free(t_trigger *x)
{
    freebytes(x->x_vec, x->x_n * sizeof(*x->x_vec));
}

static void trigger_setup(void)
{
    trigger_class = class_new(gensym("trigger"), trigger_new, trigger_free,
    	sizeof(t_trigger), 0, A_GIMME, 0);
    class_addcreator(trigger_new, gensym("t"), A_GIMME, 0);
    class_addlist(trigger_class, trigger_list);
    class_addbang(trigger_class, trigger_bang);
    class_addint(trigger_class, trigger_int);
    class_addfloat(trigger_class, (t_method)trigger_float);
    class_addsymbol(trigger_class, trigger_symbol);
}

/* -------------------------- spigot ------------------------------ */
static t_class *spigot_class;

typedef struct _spigot
{
    t_object x_obj;
    int x_state;
} t_spigot;

static void *spigot_new(t_symbol *s)
{
    t_spigot *x = (t_spigot *)pd_new(spigot_class);
    intinlet_new(&x->x_obj, &x->x_state);
    outlet_new(&x->x_obj, 0);
    x->x_state = 0;
    return (x);
}

static void spigot_bang(t_spigot *x)
{
    if (x->x_state) outlet_bang(x->x_obj.ob_outlet);
}

static void spigot_int(t_spigot *x, t_int n)
{
    if (x->x_state) outlet_int(x->x_obj.ob_outlet, n);
}

static void spigot_float(t_spigot *x, t_float f)
{
    if (x->x_state) outlet_float(x->x_obj.ob_outlet, f);
}

static void spigot_list(t_spigot *x, t_symbol *s, int argc, t_atom *argv)
{
    if (x->x_state) outlet_list(x->x_obj.ob_outlet, s, argc, argv);
}

static void spigot_anything(t_spigot *x, t_symbol *s, int argc, t_atom *argv)
{
    if (x->x_state) outlet_anything(x->x_obj.ob_outlet, s, argc, argv);
}

static void spigot_setup(void)
{
    spigot_class = class_new(gensym("spigot"), spigot_new, 0,
    	sizeof(t_spigot), 0, A_DEFSYM, 0);
    class_addbang(spigot_class, spigot_bang);
    class_addint(spigot_class, spigot_int);
    class_addfloat(spigot_class, (t_method)spigot_float);
    class_addlist(spigot_class, spigot_list);
    class_addanything(spigot_class, spigot_anything);
}

static t_class *moses_class;   	/* ----------- moses --------------- */

typedef struct _moses
{
    t_object x_ob;
    t_outlet *x_out2;
    float x_y;
} t_moses;

static void *moses_new(t_floatarg f)
{
    t_moses *x = (t_moses *)pd_new(moses_class);
    floatinlet_new(&x->x_ob, &x->x_y);
    x->x_out2 = outlet_new(&x->x_ob, &s_float);
    outlet_new(&x->x_ob, &s_float);
    x->x_y = f;
    return (x);
}

static void moses_float(t_moses *x, t_float f)
{
    if (f < x->x_y) outlet_float(x->x_ob.ob_outlet, f);
    else outlet_float(x->x_out2, f);
}

static void moses_setup(void)
{
    moses_class = class_new(gensym("moses"), (t_newmethod)moses_new, 0,
    	sizeof(t_moses), 0, A_DEFFLOAT, 0);
    class_addfloat(moses_class, (t_method)moses_float);
}

/* -------------- overall setup routine for this file ----------------- */

void x_connective_setup(void)
{
    pdint_setup();
    pdfloat_setup();
    send_setup();
    receive_setup();
    select_setup();
    route_setup();
    pack_setup();
    unpack_setup();
    trigger_setup();
    spigot_setup();
    moses_setup();
}
