/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

/* MIDI. */

#include "m_pd.h"
void outmidi_noteon(int portno, int channel, int pitch, int velo);
void outmidi_controlchange(int portno, int channel, int ctlno, int value);
void outmidi_programchange(int portno, int channel, int value);
void outmidi_pitchbend(int portno, int channel, int value);
void outmidi_aftertouch(int portno, int channel, int value);
void outmidi_polyaftertouch(int portno, int channel, int pitch, int value);

/* ----------------------- notein ------------------------- */

static t_symbol *notein_sym;

static t_class *notein_class;

typedef struct _notein
{
    t_object x_obj;
    t_int x_channel;
    t_outlet *x_outlet1;
    t_outlet *x_outlet2;
    t_outlet *x_outlet3;
} t_notein;

static void *notein_new(int n)
{
    t_notein *x = (t_notein *)pd_new(notein_class);
    x->x_channel = n;
    if (!n) x->x_outlet3 = outlet_new(&x->x_obj, &s_int);
    x->x_outlet2 = outlet_new(&x->x_obj, &s_int);
    x->x_outlet1 = outlet_new(&x->x_obj, &s_int);
    pd_bind(&x->x_obj.ob_pd, notein_sym);
    return (x);
}

static void notein_list(t_notein *x, t_symbol *s, int argc, t_atom *argv)
{
    int pitch = atom_getintarg(0, argc, argv);
    int velo = atom_getintarg(1, argc, argv);
    int channel = atom_getintarg(2, argc, argv);
    if (x->x_channel)
    {
    	if (channel != x->x_channel) return;
    	outlet_int(x->x_outlet2, velo);
    	outlet_int(x->x_outlet1, pitch);
    }
    else
    {
    	outlet_int(x->x_outlet3, channel);
    	outlet_int(x->x_outlet2, velo);
    	outlet_int(x->x_outlet1, pitch);
    }
}

static void notein_free(t_notein *x)
{
    pd_unbind(&x->x_obj.ob_pd, notein_sym);
}

static void notein_setup(void)
{
    notein_class = class_new(gensym("notein"), notein_new, notein_free,
    	sizeof(t_notein), CLASS_NOINLET, A_DEFINT, 0);
    class_addlist(notein_class, notein_list);
    notein_sym = gensym("#notein");
}

void inmidi_noteon(int portno, int channel, int pitch, int velo)
{
    if (notein_sym->s_thing)
    {
    	t_atom at[3];
    	SETINT(at, pitch);
    	SETINT(at+1, velo);
    	SETINT(at+2, (channel + (portno << 4)));
    	pd_list(notein_sym->s_thing, &s_list, 3, at);
    }
}

/* ------------------------ other incoming MIDI -- ignored for now */

void inmidi_controlchange(int portno, int channel, int ctlnumber, int value) {}
void inmidi_programchange(int portno, int channel, int value) {}
void inmidi_pitchbend(int portno, int channel, int value) {}
void inmidi_aftertouch(int portno, int channel, int value) {}
void inmidi_polyaftertouch(int portno, int channel, int pitch, int value) {}

/* -------------------------- noteout -------------------------- */

static t_class *noteout_class;

typedef struct _noteout
{
    t_object x_obj;
    t_int x_velo;
    t_int x_channel;
} t_noteout;

static void *noteout_new(int channel)
{
    t_noteout *x = (t_noteout *)pd_new(noteout_class);
    x->x_velo = 0;
    if (channel <= 0) channel = 1;
    x->x_channel = channel;
    intinlet_new(&x->x_obj, &x->x_channel);
    intinlet_new(&x->x_obj, &x->x_velo);
    return (x);
}

static void noteout_int(t_noteout *x, t_int n)
{
    int binchan = x->x_channel - 1;
    outmidi_noteon((binchan >> 4),
    	(binchan & 15), n, x->x_velo);
}

static void noteout_setup(void)
{
    noteout_class = class_new(gensym("noteout"), noteout_new, 0,
    	sizeof(t_noteout), 0, A_DEFINT, 0);
    class_addint(noteout_class, noteout_int);
}


/* -------------------------- ctlout -------------------------- */

static t_class *ctlout_class;

typedef struct _ctlout
{
    t_object x_obj;
    t_int x_ctl;
    t_int x_channel;
} t_ctlout;

static void *ctlout_new(int ctl, int channel)
{
    t_ctlout *x = (t_ctlout *)pd_new(ctlout_class);
    x->x_ctl = ctl;
    if (channel <= 0) channel = 1;
    x->x_channel = channel;
    intinlet_new(&x->x_obj, &x->x_channel);
    intinlet_new(&x->x_obj, &x->x_ctl);
    return (x);
}

static void ctlout_int(t_ctlout *x, t_int n)
{
    int binchan = x->x_channel - 1;
    outmidi_controlchange((binchan >> 4),
    	(binchan & 15), x->x_ctl, n);
}

static void ctlout_setup(void)
{
    ctlout_class = class_new(gensym("ctlout"), ctlout_new, 0,
    	sizeof(t_ctlout), 0, A_DEFINT, A_DEFINT, 0);
    class_addint(ctlout_class, ctlout_int);
}


/* -------------------------- pgmout -------------------------- */

static t_class *pgmout_class;

typedef struct _pgmout
{
    t_object x_obj;
    t_int x_channel;
} t_pgmout;

static void *pgmout_new(int channel)
{
    t_pgmout *x = (t_pgmout *)pd_new(pgmout_class);
    if (channel <= 0) channel = 1;
    x->x_channel = channel;
    intinlet_new(&x->x_obj, &x->x_channel);
    return (x);
}

static void pgmout_int(t_pgmout *x, t_int n)
{
    int binchan = x->x_channel - 1;
    n -= 1;
    if (n < 0) n = 0;
    else if (n > 127) n = 127;
    outmidi_programchange((binchan >> 4),
    	(binchan & 15), n);
}

static void pgmout_setup(void)
{
    pgmout_class = class_new(gensym("pgmout"), pgmout_new, 0,
    	sizeof(t_pgmout), 0, A_DEFINT, 0);
    class_addint(pgmout_class, pgmout_int);
}


/* -------------------------- bendout -------------------------- */

static t_class *bendout_class;

typedef struct _bendout
{
    t_object x_obj;
    t_int x_channel;
} t_bendout;

static void *bendout_new(int channel)
{
    t_bendout *x = (t_bendout *)pd_new(bendout_class);
    if (channel <= 0) channel = 1;
    x->x_channel = channel;
    intinlet_new(&x->x_obj, &x->x_channel);
    return (x);
}

static void bendout_int(t_bendout *x, t_int n)
{
    int binchan = x->x_channel - 1;
    n += 8192;
    outmidi_pitchbend((binchan >> 4), (binchan & 15), n);
}

static void bendout_setup(void)
{
    bendout_class = class_new(gensym("bendout"), bendout_new, 0,
    	sizeof(t_bendout), 0, A_DEFINT, 0);
    class_addint(bendout_class, bendout_int);
}

/* -------------------------- makenote -------------------------- */

static t_class *makenote_class;

typedef struct _hang
{
    t_clock *h_clock;
    struct _hang *h_next;
    t_int h_pitch;
    struct _makenote *h_owner;
} t_hang;

typedef struct _makenote
{
    t_object x_obj;
    t_int x_velo;
    t_int x_dur;
    t_outlet *x_pitchout;
    t_outlet *x_velout;
    t_hang *x_hang;
} t_makenote;

static void *makenote_new(int velo, int dur)
{
    t_makenote *x = (t_makenote *)pd_new(makenote_class);
    x->x_velo = velo;
    x->x_dur = dur;
    intinlet_new(&x->x_obj, &x->x_dur);
    intinlet_new(&x->x_obj, &x->x_velo);
    x->x_velout = outlet_new(&x->x_obj, &s_int);
    x->x_pitchout = outlet_new(&x->x_obj, &s_int);
    x->x_hang = 0;
    return (x);
}

static void makenote_tick(t_hang *hang)
{
    t_makenote *x = hang->h_owner;
    t_hang *h2, *h3;
    outlet_int(x->x_velout, 0);
    outlet_int(x->x_pitchout, hang->h_pitch);
    if (x->x_hang == hang) x->x_hang = hang->h_next;
    else for (h2 = x->x_hang; h3 = h2->h_next; h2 = h3)
    {
    	if (h3 == hang)
    	{
    	    h2->h_next = h3->h_next;
    	    break;
    	}
    }
    clock_free(hang->h_clock);
    freebytes(hang, sizeof(*hang));
}

static void makenote_int(t_makenote *x, t_int n)
{
    t_hang *hang;
    if (!x->x_velo) return;
    outlet_int(x->x_velout, x->x_velo);
    outlet_int(x->x_pitchout, n);
    hang = (t_hang *)getbytes(sizeof *hang);
    hang->h_next = x->x_hang;
    x->x_hang = hang;
    hang->h_pitch = n;
    hang->h_owner = x;
    hang->h_clock = clock_new(hang, makenote_tick);
    clock_delay(hang->h_clock, (x->x_dur >= 0 ? x->x_dur : 0));
}

static void makenote_stop(t_makenote *x)
{
    t_hang *hang;
    while (hang = x->x_hang)
    {
    	outlet_int(x->x_velout, 0);
    	outlet_int(x->x_pitchout, hang->h_pitch);
    	x->x_hang = hang->h_next;
    	clock_free(hang->h_clock);
    	freebytes(hang, sizeof(*hang));
    }
}

static void makenote_clear(t_makenote *x)
{
    t_hang *hang;
    while (hang = x->x_hang)
    {
    	x->x_hang = hang->h_next;
    	clock_free(hang->h_clock);
    	freebytes(hang, sizeof(*hang));
    }
}

static void makenote_setup(void)
{
    makenote_class = class_new(gensym("makenote"), makenote_new, makenote_clear,
    	sizeof(t_makenote), 0, A_DEFINT, A_DEFINT, 0);
    class_addint(makenote_class, makenote_int);
    class_addmethod(makenote_class, makenote_stop, gensym("stop"), 0);
    class_addmethod(makenote_class, makenote_clear, gensym("clear"), 0);
}

/* -------------------------- stripnote -------------------------- */

static t_class *stripnote_class;

typedef struct _stripnote
{
    t_object x_obj;
    t_int x_velo;
    t_outlet *x_pitchout;
    t_outlet *x_velout;
} t_stripnote;

static void *stripnote_new()
{
    t_stripnote *x = (t_stripnote *)pd_new(stripnote_class);
    intinlet_new(&x->x_obj, &x->x_velo);
    x->x_velout = outlet_new(&x->x_obj, &s_int);
    x->x_pitchout = outlet_new(&x->x_obj, &s_int);
    return (x);
}
    
static void stripnote_int(t_stripnote *x, t_int n)
{
    t_hang *hang;
    if (!x->x_velo) return;
    outlet_int(x->x_velout, x->x_velo);
    outlet_int(x->x_pitchout, n);
}

static void stripnote_setup(void)
{
    stripnote_class = class_new(gensym("stripnote"), stripnote_new, 0,
    	sizeof(t_stripnote), 0, 0);
    class_addint(stripnote_class, stripnote_int);
}

void x_midi_setup(void)
{
    notein_setup();
    noteout_setup();
    ctlout_setup();
    pgmout_setup();
    bendout_setup();
    makenote_setup();
    stripnote_setup();
}
